import os
import platform
import re
import textwrap

import pytest

from conan.test.assets.cmake import gen_cmakelists
from conan.test.assets.sources import gen_function_cpp
from conan.test.utils.scm import create_local_git_repo, git_add_changes_commit, git_create_bare_repo
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient
from conan.internal.util.files import rmdir, save_files, save


@pytest.mark.tool("git")
class TestGitBasicCapture:
    """ base Git capture operations. They do not raise (unless errors)
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.scm import Git

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, self.recipe_folder, excluded=["myfile.txt", "mynew.txt", "file with spaces.txt"])
                commit = git.get_commit()
                repo_commit = git.get_commit(repository=True)
                url = git.get_remote_url()
                self.output.info("URL: {}".format(url))
                self.output.info("COMMIT: {}".format(commit))
                self.output.info("REPO_COMMIT: {}".format(repo_commit))
                in_remote = git.commit_in_remote(commit)
                self.output.info("COMMIT IN REMOTE: {}".format(in_remote))
                self.output.info("DIRTY: {}".format(git.is_dirty()))
        """)

    def test_capture_commit_local(self):
        """
        A local repo, without remote, will have commit, but no URL
        """
        c = TestClient()
        c.save({"conanfile.py": self.conanfile})
        commit = c.init_git_repo()
        c.run("export .")
        assert "pkg/0.1: COMMIT: {}".format(commit) in c.out
        assert "pkg/0.1: URL: None" in c.out
        assert "pkg/0.1: COMMIT IN REMOTE: False" in c.out
        assert "pkg/0.1: DIRTY: False" in c.out

    def test_capture_remote_url(self):
        """
        a cloned repo, will have a default "origin" remote and will manage to get URL
        """
        folder = temp_folder()
        url, commit = create_local_git_repo(files={"conanfile.py": self.conanfile}, folder=folder)

        c = TestClient()
        c.run_command('git clone "{}" myclone'.format(folder))
        with c.chdir("myclone"):
            c.run("export .")
            assert "pkg/0.1: COMMIT: {}".format(commit) in c.out
            assert "pkg/0.1: REPO_COMMIT: {}".format(commit) in c.out
            assert "pkg/0.1: URL: {}".format(url) in c.out
            assert "pkg/0.1: COMMIT IN REMOTE: True" in c.out
            assert "pkg/0.1: DIRTY: False" in c.out

    def test_capture_remote_pushed_commit(self):
        """
        a cloned repo, after doing some new commit, no longer commit in remote, until push
        """
        url = git_create_bare_repo()

        c = TestClient()
        c.run_command('git clone "{}" myclone'.format(url))
        with c.chdir("myclone"):
            c.save({"conanfile.py": self.conanfile + "\n# some coment!"})
            new_commit = git_add_changes_commit(c.current_folder)

            c.run("export .")
            assert "pkg/0.1: COMMIT: {}".format(new_commit) in c.out
            assert "pkg/0.1: URL: {}".format(url) in c.out
            assert "pkg/0.1: COMMIT IN REMOTE: False" in c.out
            assert "pkg/0.1: DIRTY: False" in c.out
            c.run_command("git push")
            c.run("export .")
            assert "pkg/0.1: COMMIT: {}".format(new_commit) in c.out
            assert "pkg/0.1: URL: {}".format(url) in c.out
            assert "pkg/0.1: COMMIT IN REMOTE: True" in c.out
            assert "pkg/0.1: DIRTY: False" in c.out

    def test_capture_commit_local_subfolder(self):
        """
        A local repo, without remote, will have commit, but no URL, and sibling folders
        can be dirty, no prob
        """
        c = TestClient()
        c.save({"subfolder/conanfile.py": self.conanfile,
                "other/myfile.txt": "content"})
        commit = c.init_git_repo()
        c.save({"other/myfile.txt": "change content"})
        c.run("export subfolder")
        assert "pkg/0.1: COMMIT: {}".format(commit) in c.out
        assert "pkg/0.1: REPO_COMMIT: {}".format(commit) in c.out
        assert "pkg/0.1: URL: None" in c.out
        assert "pkg/0.1: COMMIT IN REMOTE: False" in c.out
        assert "pkg/0.1: DIRTY: False" in c.out
        commit2 = git_add_changes_commit(c.current_folder, msg="fix")
        c.run("export subfolder")
        assert "pkg/0.1: COMMIT: {}".format(commit) in c.out
        assert "pkg/0.1: REPO_COMMIT: {}".format(commit2) in c.out
        assert "pkg/0.1: URL: None" in c.out
        assert "pkg/0.1: COMMIT IN REMOTE: False" in c.out
        assert "pkg/0.1: DIRTY: False" in c.out

    def test_git_excluded(self):
        """
        A local repo, without remote, will have commit, but no URL
        """
        c = TestClient()
        c.save({"conanfile.py": self.conanfile,
                "myfile.txt": ""})
        c.init_git_repo()
        c.run("export . -vvv")
        assert "pkg/0.1: DIRTY: False" in c.out
        c.save({"myfile.txt": "changed",
                "mynew.txt": "new",
                "file with spaces.txt": "hello"})
        c.run("export .")
        assert "pkg/0.1: DIRTY: False" in c.out
        c.save({"other.txt": "new"})
        c.run("export .")
        assert "pkg/0.1: DIRTY: True" in c.out

        conf_excluded = f'core.scm:excluded+=["other.txt"]'
        save(c.paths.global_conf_path, conf_excluded)
        c.run("export .")
        assert "pkg/0.1: DIRTY: False" in c.out


@pytest.mark.tool("git")
class TestGitCaptureSCM:
    """ test the get_url_and_commit() high level method intended for SCM capturing
    into conandata.yaml
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.scm import Git

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, self.recipe_folder)
                scm_url, scm_commit = git.get_url_and_commit()
                self.output.info("SCM URL: {}".format(scm_url))
                self.output.info("SCM COMMIT: {}".format(scm_commit))
        """)

    def test_capture_commit_local(self):
        """
        A local repo, without remote, will provide its own URL to the export(),
        and if it has local changes, it will be marked as dirty, and raise an error
        """
        c = TestClient()
        c.save({"conanfile.py": self.conanfile})
        commit = c.init_git_repo()
        c.run("export .")
        assert "This revision will not be buildable in other computer" in c.out
        assert "pkg/0.1: SCM COMMIT: {}".format(commit) in c.out
        assert "pkg/0.1: SCM URL: {}".format(c.current_folder.replace("\\", "/")) in c.out

        c.save({"conanfile.py": self.conanfile + "\n# something...."})
        c.run("export .", assert_error=True)
        assert "Repo is dirty, cannot capture url and commit" in c.out

    def test_capture_commit_local_repository(self):
        """
        same as above, but with ``get_url_and_commit(repository=True)``
        """
        c = TestClient()
        c.save({"pkg/conanfile.py": self.conanfile.replace("get_url_and_commit()",
                                                           "get_url_and_commit(repository=True)"),
                "somefile.txt": ""})
        commit = c.init_git_repo()
        c.run("export pkg")
        assert "This revision will not be buildable in other computer" in c.out
        assert "pkg/0.1: SCM COMMIT: {}".format(commit) in c.out
        assert "pkg/0.1: SCM URL: {}".format(c.current_folder.replace("\\", "/")) in c.out

        c.save({"somefile.txt": "something"})
        c.run("export pkg", assert_error=True)
        assert "Repo is dirty, cannot capture url and commit" in c.out

    def test_capture_remote_url(self):
        """
        a cloned repo that is expored, will report the URL of the remote
        """
        folder = temp_folder()
        url, commit = create_local_git_repo(files={"conanfile.py": self.conanfile}, folder=folder)

        c = TestClient()
        c.run_command('git clone "{}" myclone'.format(folder))
        with c.chdir("myclone"):
            c.run("export .")
            assert "pkg/0.1: SCM COMMIT: {}".format(commit) in c.out
            assert "pkg/0.1: SCM URL: {}".format(url) in c.out

    def test_capture_remote_pushed_commit(self):
        """
        a cloned repo, after doing some new commit, no longer commit in remote, until push
        """
        url = git_create_bare_repo()

        c = TestClient()
        c.run_command('git clone "{}" myclone'.format(url))
        with c.chdir("myclone"):
            c.save({"conanfile.py": self.conanfile + "\n# some coment!"})
            new_commit = git_add_changes_commit(c.current_folder)

            c.run("export .")
            assert "This revision will not be buildable in other computer" in c.out
            assert "pkg/0.1: SCM COMMIT: {}".format(new_commit) in c.out
            # NOTE: commit not pushed yet, so locally is the current folder
            assert "pkg/0.1: SCM URL: {}".format(c.current_folder.replace("\\", "/")) in c.out
            c.run_command("git push")
            c.run("export .")
            assert "pkg/0.1: SCM COMMIT: {}".format(new_commit) in c.out
            assert "pkg/0.1: SCM URL: {}".format(url) in c.out

    def test_capture_commit_modified_config(self):
        """
        A clean repo with the status.branch git config set to on
        Expected to not raise an error an return the commit and url
        """
        folder = temp_folder()
        url, commit = create_local_git_repo(files={"conanfile.py": self.conanfile}, folder=folder)
        c = TestClient()
        with c.chdir(folder):
            c.run_command("git config --local status.branch true")
            c.run("export .")
            assert "pkg/0.1: SCM COMMIT: {}".format(commit) in c.out
            assert "pkg/0.1: SCM URL: {}".format(url) in c.out

    def test_capture_commit_modified_config_untracked(self):
        """
        A dirty repo with the showUntrackedFiles git config set to off.
        Expected to throw an exception
        """
        folder = temp_folder()
        create_local_git_repo(files={"conanfile.py": self.conanfile}, folder=folder)
        c = TestClient()
        with c.chdir(folder):
            c.save(files={"some_header.h": "now the repo is dirty"})
            c.run_command("git config --local status.showUntrackedFiles no")
            c.run("export .", assert_error=True)
            assert "Repo is dirty, cannot capture url and commit" in c.out


@pytest.mark.tool("git")
class TestGitBasicClone:
    """ base Git cloning operations
    """
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def layout(self):
                self.folders.source = "source"

            def source(self):
                git = Git(self)
                git.clone(url="{url}", target=".")
                git.checkout(commit="{commit}")
                self.output.info("MYCMAKE: {{}}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {{}}".format(load(self, "src/myfile.h")))
        """)

    def test_clone_checkout(self):
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder)
        # This second commit will NOT be used, as I will use the above commit in the conanfile
        save_files(path=folder, files={"src/myfile.h": "my2header2!"})
        git_add_changes_commit(folder=folder)

        c = TestClient()
        c.save({"conanfile.py": self.conanfile.format(url=url, commit=commit)})
        c.run("create . -v")
        # Clone is not printed, it might contain tokens
        assert 'pkg/0.1: RUN: git clone "<hidden>"  "."' in c.out
        assert "pkg/0.1: RUN: git checkout" in c.out
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out

        # It also works in local flow
        c.run("source .")
        assert "conanfile.py (pkg/0.1): MYCMAKE: mycmake" in c.out
        assert "conanfile.py (pkg/0.1): MYFILE: myheader!" in c.out
        assert c.load("source/src/myfile.h") == "myheader!"
        assert c.load("source/CMakeLists.txt") == "mycmake"

    def test_clone_url_not_hidden(self):
        conanfile = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.scm import Git
            from conan.tools.files import load

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"

                def layout(self):
                    self.folders.source = "source"

                def source(self):
                    git = Git(self)
                    git.clone(url="{url}", target=".", hide_url=False)
            """)
        folder = os.path.join(temp_folder(), "myrepo")
        url, _ = create_local_git_repo(files={"CMakeLists.txt": "mycmake"}, folder=folder)

        c = TestClient(light=True)
        c.save({"conanfile.py": conanfile.format(url=url)})
        c.run("create . -v")
        # Clone URL is explicitly printed
        assert f'pkg/0.1: RUN: git clone "{url}"  "."' in c.out

        # It also works in local flow
        c.run("source .")
        assert f'conanfile.py (pkg/0.1): RUN: git clone "{url}"  "."' in c.out

    def test_clone_target(self):
        # Clone to a different target folder
        # https://github.com/conan-io/conan/issues/14058
        conanfile = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.scm import Git
            from conan.tools.files import load

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"

                def layout(self):
                    self.folders.source = "source"

                def source(self):
                    # Alternative, first defining the folder
                    # git = Git(self, "target")
                    # git.clone(url="{url}", target=".")
                    # git.checkout(commit="{commit}")

                    git = Git(self)
                    git.clone(url="{url}", target="tar get") # git clone url target
                    git.folder = "tar get"                   # cd target
                    git.checkout(commit="{commit}")         # git checkout commit

                    self.output.info("MYCMAKE: {{}}".format(load(self, "tar get/CMakeLists.txt")))
                    self.output.info("MYFILE: {{}}".format(load(self, "tar get/src/myfile.h")))
                """)
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder)
        # This second commit will NOT be used, as I will use the above commit in the conanfile
        save_files(path=folder, files={"src/myfile.h": "my2header2!"})
        git_add_changes_commit(folder=folder)

        c = TestClient()
        c.save({"conanfile.py": conanfile.format(url=url, commit=commit)})
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out

    @pytest.mark.tool("msys2")
    def test_clone_msys2_win_bash(self):
        # To avoid regression in https://github.com/conan-io/conan/issues/14754
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder)

        c = TestClient()
        conanfile_win_bash = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.scm import Git
            from conan.tools.files import load

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                win_bash = True

                def layout(self):
                    self.folders.source = "source"

                def source(self):
                    git = Git(self)
                    git.clone(url="{url}", target=".")
                    git.checkout(commit="{commit}")
                    self.output.info("MYCMAKE: {{}}".format(load(self, "CMakeLists.txt")))
                    self.output.info("MYFILE: {{}}".format(load(self, "src/myfile.h")))
            """)
        c.save({"conanfile.py": conanfile_win_bash.format(url=url, commit=commit)})
        conf = "-c tools.microsoft.bash:subsystem=msys2 -c tools.microsoft.bash:path=bash.exe"
        c.run(f"create . {conf}")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out

        # It also works in local flow, not running in msys2 at all
        c.run(f"source .")
        assert "conanfile.py (pkg/0.1): MYCMAKE: mycmake" in c.out
        assert "conanfile.py (pkg/0.1): MYFILE: myheader!" in c.out
        assert c.load("source/src/myfile.h") == "myheader!"
        assert c.load("source/CMakeLists.txt") == "mycmake"


@pytest.mark.tool("git")
class TestGitShallowClone:
    """ base Git cloning operations
    """
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def layout(self):
                self.folders.source = "source"

            def source(self):
                git = Git(self)
                git.fetch_commit(url="{url}", commit="{commit}")
                self.output.info("MYCMAKE: {{}}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {{}}".format(load(self, "src/myfile.h")))
        """)

    @pytest.mark.skipif(platform.system() == "Linux", reason="Git version in Linux not support it")
    def test_clone_checkout(self):
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder)
        # This second commit will NOT be used, as I will use the above commit in the conanfile
        save_files(path=folder, files={"src/myfile.h": "my2header2!"})
        git_add_changes_commit(folder=folder)

        c = TestClient()
        c.save({"conanfile.py": self.conanfile.format(url=url, commit=commit)})
        c.run("create . -v")
        assert 'pkg/0.1: RUN: git remote add origin "<hidden>"' in c.out
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out

        # It also works in local flow
        c.run("source .")
        assert "conanfile.py (pkg/0.1): MYCMAKE: mycmake" in c.out
        assert "conanfile.py (pkg/0.1): MYFILE: myheader!" in c.out
        assert c.load("source/src/myfile.h") == "myheader!"
        assert c.load("source/CMakeLists.txt") == "mycmake"

    def test_clone_url_not_hidden(self):
        conanfile = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.scm import Git
            from conan.tools.files import load

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"

                def layout(self):
                    self.folders.source = "source"

                def source(self):
                    git = Git(self)
                    git.fetch_commit(url="{url}", commit="{commit}", hide_url=False)
            """)
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"CMakeLists.txt": "mycmake"}, folder=folder)

        c = TestClient(light=True)
        c.save({"conanfile.py": conanfile.format(url=url, commit=commit)})
        c.run("create . -v")
        # Clone URL is explicitly printed
        assert f'pkg/0.1: RUN: git remote add origin "{url}"' in c.out

        # It also works in local flow
        c.run("source .")
        assert f'conanfile.py (pkg/0.1): RUN: git remote add origin "{url}"' in c.out

    @pytest.mark.skipif(platform.system() == "Linux", reason="Git version in Linux not support it")
    def test_clone_to_subfolder(self):
        conanfile = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.scm import Git
            from conan.tools.files import load

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"

                def layout(self):
                    self.folders.source = "source"

                def source(self):
                    git = Git(self, folder="folder")
                    git.fetch_commit(url="{url}", commit="{commit}")
                    self.output.info("MYCMAKE: {{}}".format(load(self, "folder/CMakeLists.txt")))
                    self.output.info("MYFILE: {{}}".format(load(self, "folder/src/myfile.h")))
            """)
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder)
        # This second commit will NOT be used, as I will use the above commit in the conanfile
        save_files(path=folder, files={"src/myfile.h": "my2header2!"})
        git_add_changes_commit(folder=folder)

        c = TestClient()
        c.save({"conanfile.py": conanfile.format(url=url, commit=commit)})
        c.run("create . -v")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out

        # It also works in local flow
        c.run("source .")
        assert "conanfile.py (pkg/0.1): MYCMAKE: mycmake" in c.out
        assert "conanfile.py (pkg/0.1): MYFILE: myheader!" in c.out
        assert c.load("source/folder/CMakeLists.txt") == "mycmake"
        assert c.load("source/folder/src/myfile.h") == "myheader!"

class TestGitCloneWithArgs:
    """ Git cloning passing additional arguments
    """
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def layout(self):
                self.folders.source = "source"

            def source(self):
                git = Git(self)
                git.clone(url="{url}", target=".", args={args})
                self.output.info("MYCMAKE: {{}}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {{}}".format(load(self, "src/myfile.h")))
        """)

    def test_clone_specify_branch_or_tag(self):
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder,
                                            commits=3, branch="main", tags=["v1.2.3"])

        c = TestClient()
        git_args = ['--branch', 'main']
        c.save({"conanfile.py": self.conanfile.format(url=url, commit=commit, args=str(git_args))})
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out

        git_args = ['--branch', 'v1.2.3']
        c.save({"conanfile.py": self.conanfile.format(url=url, commit=commit, args=str(git_args))})
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out

    def test_clone_invalid_branch_argument(self):
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder,
                                            commits=3, branch="main", tags=["v1.2.3"])
        c = TestClient()
        git_args = ['--branch', 'foobar']
        c.save({"conanfile.py": self.conanfile.format(url=url, commit=commit, args=str(git_args))})
        c.run("create .", assert_error=True)
        assert "Remote branch foobar not found" in c.out


@pytest.mark.tool("git")
class TestGitBasicSCMFlow:
    """ Build the full new SCM approach:
    - export() captures the URL and commit with get_url_and_commit(
    - export() stores it in conandata.yml
    - source() recovers the info from conandata.yml and clones it
    """
    conanfile_full = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load, update_conandata

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, self.recipe_folder)
                scm_url, scm_commit = git.get_url_and_commit()
                update_conandata(self, {"sources": {"commit": scm_commit, "url": scm_url}})

            def layout(self):
                self.folders.source = "."

            def source(self):
                git = Git(self)
                sources = self.conan_data["sources"]
                git.clone(url=sources["url"], target=".")
                git.checkout(commit=sources["commit"])
                self.output.info("MYCMAKE: {}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {}".format(load(self, "src/myfile.h")))

            def build(self):
                cmake = os.path.join(self.source_folder, "CMakeLists.txt")
                file_h = os.path.join(self.source_folder, "src/myfile.h")
                self.output.info("MYCMAKE-BUILD: {}".format(load(self, cmake)))
                self.output.info("MYFILE-BUILD: {}".format(load(self, file_h)))
        """)
    conanfile_scm = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load, trim_conandata

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                Git(self).coordinates_to_conandata()
                trim_conandata(self)  # to test it does not affect

            def layout(self):
                self.folders.source = "."

            def source(self):
                Git(self).checkout_from_conandata_coordinates()
                self.output.info("MYCMAKE: {}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {}".format(load(self, "src/myfile.h")))

            def build(self):
                cmake = os.path.join(self.source_folder, "CMakeLists.txt")
                file_h = os.path.join(self.source_folder, "src/myfile.h")
                self.output.info("MYCMAKE-BUILD: {}".format(load(self, cmake)))
                self.output.info("MYFILE-BUILD: {}".format(load(self, file_h)))
        """)

    @pytest.mark.parametrize("conanfile_scm", [False, True])
    def test_full_scm(self, conanfile_scm):
        conanfile = self.conanfile_scm if conanfile_scm else self.conanfile_full
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"conanfile.py": conanfile,
                                                   "src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder)

        c = TestClient(default_server_user=True)
        c.run_command('git clone "file://{}" .'.format(url))
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out
        c.run("upload * -c -r=default")

        # Do a change and commit, this commit will not be used by package
        save_files(path=folder, files={"src/myfile.h": "my2header2!"})
        git_add_changes_commit(folder=folder)

        # use another fresh client
        c2 = TestClient(servers=c.servers)
        c2.run("install --requires=pkg/0.1@ --build=pkg*")
        assert "pkg/0.1: MYCMAKE: mycmake" in c2.out
        assert "pkg/0.1: MYFILE: myheader!" in c2.out

        # local flow
        c.run("install .")
        c.run("build .")
        assert "conanfile.py (pkg/0.1): MYCMAKE-BUILD: mycmake" in c.out
        assert "conanfile.py (pkg/0.1): MYFILE-BUILD: myheader!" in c.out

    @pytest.mark.parametrize("conanfile_scm", [False, True])
    def test_branch_flow(self, conanfile_scm):
        """ Testing that when a user creates a branch, and pushes a commit,
        the package can still be built from sources, and get_url_and_commit() captures the
        remote URL and not the local
        """
        conanfile = self.conanfile_scm if conanfile_scm else self.conanfile_full
        url = git_create_bare_repo()
        c = TestClient(default_server_user=True)
        c.run_command('git clone "file://{}" .'.format(url))
        c.save({"conanfile.py": conanfile,
                "src/myfile.h": "myheader!",
                "CMakeLists.txt": "mycmake"})
        c.run_command("git checkout -b mybranch")
        git_add_changes_commit(folder=c.current_folder)
        c.run_command("git push --set-upstream origin mybranch")
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out
        c.run("upload * -c -r=default")
        rmdir(c.current_folder)  # Remove current folder to make sure things are not used from here

        # use another fresh client
        c2 = TestClient(servers=c.servers)
        c2.run("install --requires=pkg/0.1@ --build=pkg*")
        assert "pkg/0.1: MYCMAKE: mycmake" in c2.out
        assert "pkg/0.1: MYFILE: myheader!" in c2.out

    def test_fetch_commit(self):
        """ Testing fetch commit
        """
        url = git_create_bare_repo()
        c = TestClient(default_server_user=True)
        c.run_command('git clone "file://{}" .'.format(url))
        c.save({"conanfile.py": self.conanfile_scm,
                "src/myfile.h": "myheader!",
                "CMakeLists.txt": "mycmake"})
        c.run_command("git checkout -b mybranch")
        git_add_changes_commit(folder=c.current_folder)
        c.run_command("git push --set-upstream origin mybranch")
        c.run("create .")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out
        c.run("upload * -c -r=default")
        # Create an orphan commit by removing the branch
        c.run_command("git push origin --delete mybranch")
        rmdir(c.current_folder)  # Remove current folder to make sure things are not used from here

        # use another fresh client
        c2 = TestClient(servers=c.servers)
        c2.run("install --requires=pkg/0.1@ --build=pkg*")
        assert "pkg/0.1: MYCMAKE: mycmake" in c2.out
        assert "pkg/0.1: MYFILE: myheader!" in c2.out

    def test_grafted_commit(self):
        # https://github.com/conan-io/conan/issues/18295
        folder = os.path.join(temp_folder(), "myrepo")
        conanfile = textwrap.dedent("""\
           from conan import ConanFile
           from conan.tools.scm import Git

           class ConanRecipe(ConanFile):
               name = "conan-grafted"
               version = "1.2.3"

               def export(self):
                   git = Git(self, self.recipe_folder)
                   git.coordinates_to_conandata(repository=True)
           """)
        url, commit = create_local_git_repo(files={"conanfile.py": conanfile}, folder=folder)

        c = TestClient()
        c.save({"README.txt": "my readme!"}, path=folder)
        c.run_command("git checkout -b mybranch", cwd=folder)
        orphan_commit = git_add_changes_commit(folder)
        c.run_command(f"git checkout {orphan_commit}", cwd=folder)
        c.run_command("git branch -D mybranch", cwd=folder)

        # Now the client does the clone and export
        c.run_command(f'git clone "file://{url}" . ')
        c.run_command("git --no-pager log --decorate")
        assert "grafted" not in c.out
        c.run("export .")
        c.run_command("git --no-pager log --decorate")
        assert "grafted" not in c.out


@pytest.mark.tool("git")
class TestGitBasicSCMFlowSubfolder:
    """ Same as above, but conanfile.py put in "conan" subfolder in the root
    """
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load, update_conandata

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, os.path.dirname(self.recipe_folder)) # PARENT!
                scm_url, scm_commit = git.get_url_and_commit()
                update_conandata(self, {"sources": {"commit": scm_commit, "url": scm_url}})

            def layout(self):
                self.folders.root = ".."
                self.folders.source = "."

            def source(self):
                git = Git(self)
                sources = self.conan_data["sources"]
                git.clone(url=sources["url"], target=".")
                git.checkout(commit=sources["commit"])
                self.output.info("MYCMAKE: {}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {}".format(load(self, "src/myfile.h")))

            def build(self):
                cmake = os.path.join(self.source_folder, "CMakeLists.txt")
                file_h = os.path.join(self.source_folder, "src/myfile.h")
                self.output.info("MYCMAKE-BUILD: {}".format(load(self, cmake)))
                self.output.info("MYFILE-BUILD: {}".format(load(self, file_h)))
        """)

    def test_full_scm(self):
        folder = os.path.join(temp_folder(), "myrepo")
        url, commit = create_local_git_repo(files={"conan/conanfile.py": self.conanfile,
                                                   "src/myfile.h": "myheader!",
                                                   "CMakeLists.txt": "mycmake"}, folder=folder)

        c = TestClient(default_server_user=True)
        c.run_command('git clone "{}" .'.format(url))
        c.run("create conan")
        assert "pkg/0.1: MYCMAKE: mycmake" in c.out
        assert "pkg/0.1: MYFILE: myheader!" in c.out
        c.run("upload * -c -r=default")

        # Do a change and commit, this commit will not be used by package
        save_files(path=folder, files={"src/myfile.h": "my2header2!"})
        git_add_changes_commit(folder=folder)

        # use another fresh client
        c2 = TestClient(servers=c.servers)
        c2.run("install --requires=pkg/0.1@ --build=pkg*")
        assert "pkg/0.1: MYCMAKE: mycmake" in c2.out
        assert "pkg/0.1: MYFILE: myheader!" in c2.out

        # local flow
        c.run("install conan")
        c.run("build conan")
        assert "conanfile.py (pkg/0.1): MYCMAKE-BUILD: mycmake" in c.out
        assert "conanfile.py (pkg/0.1): MYFILE-BUILD: myheader!" in c.out


@pytest.mark.tool("git")
class TestGitMonorepoSCMFlow:
    """ Build the full new SCM approach:
    Same as above but with a monorepo with multiple subprojects
    """
    # TODO: swap_child_folder() not documented, not public usage
    conanfile = textwrap.dedent("""
        import os, shutil
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import load, update_conandata, move_folder_contents

        class Pkg(ConanFile):
            name = "{pkg}"
            version = "0.1"

            {requires}

            def export(self):
                git = Git(self, self.recipe_folder)
                scm_url, scm_commit = git.get_url_and_commit()
                self.output.info("CAPTURING COMMIT: {{}}!!!".format(scm_commit))
                folder = os.path.basename(self.recipe_folder)
                update_conandata(self, {{"sources": {{"commit": scm_commit, "url": scm_url,
                                                      "folder": folder}}}})

            def layout(self):
                self.folders.source = "."

            def source(self):
                git = Git(self)
                sources = self.conan_data["sources"]
                git.clone(url=sources["url"], target=".")
                git.checkout(commit=sources["commit"])
                move_folder_contents(self, os.path.join(self.source_folder, sources["folder"]),
                                    self.source_folder)

            def build(self):
                cmake = os.path.join(self.source_folder, "CMakeLists.txt")
                file_h = os.path.join(self.source_folder, "src/myfile.h")
                self.output.info("MYCMAKE-BUILD: {{}}".format(load(self, cmake)))
                self.output.info("MYFILE-BUILD: {{}}".format(load(self, file_h)))
        """)

    def test_full_scm(self):
        folder = os.path.join(temp_folder(), "myrepo")
        conanfile1 = self.conanfile.format(pkg="pkg1", requires="")
        conanfile2 = self.conanfile.format(pkg="pkg2", requires="requires = 'pkg1/0.1'")
        url, commit = create_local_git_repo(files={"sub1/conanfile.py": conanfile1,
                                                   "sub1/src/myfile.h": "myheader1!",
                                                   "sub1/CMakeLists.txt": "mycmake1!",
                                                   "sub2/conanfile.py": conanfile2,
                                                   "sub2/src/myfile.h": "myheader2!",
                                                   "sub2/CMakeLists.txt": "mycmake2!"
                                                   },
                                            folder=folder)

        c = TestClient(default_server_user=True)
        c.run_command('git clone "{}" .'.format(url))
        c.run("create sub1")
        commit = re.search(r"CAPTURING COMMIT: (\S+)!!!", str(c.out)).group(1)
        assert "pkg1/0.1: MYCMAKE-BUILD: mycmake1!" in c.out
        assert "pkg1/0.1: MYFILE-BUILD: myheader1!" in c.out

        c.save({"sub2/src/myfile.h": "my2header!"})
        git_add_changes_commit(folder=c.current_folder)
        c.run("create sub2")
        assert "pkg2/0.1: MYCMAKE-BUILD: mycmake2!" in c.out
        assert "pkg2/0.1: MYFILE-BUILD: my2header!" in c.out

        # Exporting again sub1, gives us exactly the same revision as before
        c.run("export sub1")
        assert "CAPTURING COMMIT: {}".format(commit) in c.out
        c.run("upload * -c -r=default")

        # use another fresh client
        c2 = TestClient(servers=c.servers)
        c2.run("install --requires=pkg2/0.1@ --build=*")
        assert "pkg1/0.1: Checkout: {}".format(commit) in c2.out
        assert "pkg1/0.1: MYCMAKE-BUILD: mycmake1!" in c2.out
        assert "pkg1/0.1: MYFILE-BUILD: myheader1!" in c2.out
        assert "pkg2/0.1: MYCMAKE-BUILD: mycmake2!" in c2.out
        assert "pkg2/0.1: MYFILE-BUILD: my2header!" in c2.out

    @pytest.mark.tool("cmake")
    def test_exports_sources_common_code_layout(self):
        """ This is a copy of test_exports_sources_common_code_layout in test_in_subfolder.py
        but instead of using "exports", trying to implement it with Git features
        """
        c = TestClient()
        conanfile = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.cmake import cmake_layout, CMake
            from conan.tools.files import load, copy, save, update_conandata, move_folder_contents
            from conan.tools.scm import Git

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "os", "compiler", "build_type", "arch"
                generators = "CMakeToolchain"

                def export(self):
                    git = Git(self)
                    scm_url, scm_commit = git.get_url_and_commit()
                    update_conandata(self, {"sources": {"commit": scm_commit, "url": scm_url}})

                def layout(self):
                    self.folders.root = ".."
                    self.folders.subproject = "pkg"
                    cmake_layout(self)

                def source(self):
                    git = Git(self)
                    sources = self.conan_data["sources"]
                    git.clone(url=sources["url"], target=".")
                    git.checkout(commit=sources["commit"])
                    # Layout is pkg/pkg/<files> and pkg/common/<files>
                    # Final we want is pkg/<files> and common/<files>
                    # NOTE: This abs_path is IMPORTANT to avoid the trailing "."
                    src_folder = os.path.abspath(self.source_folder)
                    move_folder_contents(self, src_folder, os.path.dirname(src_folder))

                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
                    cmake.build()
                    self.run(os.path.join(self.cpp.build.bindirs[0], "myapp"))
                """)
        cmake_include = "include(${CMAKE_CURRENT_LIST_DIR}/../common/myutils.cmake)"
        c.save({"pkg/conanfile.py": conanfile,
                "pkg/app.cpp": gen_function_cpp(name="main", includes=["../common/myheader"],
                                                preprocessor=["MYDEFINE"]),
                "pkg/CMakeLists.txt": gen_cmakelists(appsources=["app.cpp"],
                                                     custom_content=cmake_include),
                "common/myutils.cmake": 'message(STATUS "MYUTILS.CMAKE!")',
                "common/myheader.h": '#define MYDEFINE "MYDEFINEVALUE"'})
        c.init_git_repo()

        c.run("create pkg")
        assert "MYUTILS.CMAKE!" in c.out
        assert "main: Release!" in c.out
        assert "MYDEFINE: MYDEFINEVALUE" in c.out

        # Local flow
        c.run("install pkg")
        c.run("build pkg")
        assert "MYUTILS.CMAKE!" in c.out
        assert "main: Release!" in c.out
        assert "MYDEFINE: MYDEFINEVALUE" in c.out

        c.run("install pkg -s build_type=Debug")
        c.run("build pkg -s build_type=Debug")
        assert "MYUTILS.CMAKE!" in c.out
        assert "main: Debug!" in c.out
        assert "MYDEFINE: MYDEFINEVALUE" in c.out


class TestConanFileSubfolder:
    """verify that we can have a conanfile in a subfolder
        # https://github.com/conan-io/conan/issues/11275
    """

    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.scm import Git
        from conan.tools.files import update_conandata, load

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, os.path.dirname(self.recipe_folder))
                url, commit = git.get_url_and_commit()
                # We store the current url and commit in conandata.yml
                update_conandata(self, {"sources": {"commit": commit, "url": url}})
                self.output.info("URL: {}".format(url))
                self.output.info("COMMIT: {}".format(commit))

            def layout(self):
                pass # self.folders.source = "source"

            def source(self):
                git = Git(self)
                sources = self.conan_data["sources"]
                url = sources["url"]
                commit = sources["commit"]
                git.clone(url=url, target=".")
                git.checkout(commit=commit)
                self.output.info("MYCMAKE: {}".format(load(self, "CMakeLists.txt")))
                self.output.info("MYFILE: {}".format(load(self, "src/myfile.h")))
        """)

    def test_conanfile_subfolder(self):
        """
        A local repo, without remote, will have commit, but no URL
        """
        c = TestClient()
        c.save({"conan/conanfile.py": self.conanfile,
                "CMakeLists.txt": "mycmakelists",
                "src/myfile.h": "myheader"})
        commit = c.init_git_repo()
        c.run("export conan")
        assert "pkg/0.1: COMMIT: {}".format(commit) in c.out
        assert "pkg/0.1: URL: {}".format(c.current_folder.replace("\\", "/")) in c.out

        c.run("create conan")
        assert "pkg/0.1: MYCMAKE: mycmakelists" in c.out
        assert "pkg/0.1: MYFILE: myheader" in c.out

    def test_git_run(self):
        conanfile = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.scm import Git

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                def export(self):
                    git = Git(self)
                    self.output.info(git.run("--version"))
            """)

        c = TestClient()
        c.save({"conan/conanfile.py": conanfile})
        c.init_git_repo()
        c.run("export conan")
        assert "pkg/0.1: git version" in c.out


class TestGitIncluded:
    def test_git_included(self):
        conanfile = textwrap.dedent("""
            import os
            import shutil
            from conan import ConanFile
            from conan.tools.scm import Git

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"

                def export_sources(self):
                    git = Git(self)
                    included = git.included_files()
                    for i in included:
                        dst =  os.path.join(self.export_sources_folder, i)
                        os.makedirs(os.path.dirname(dst), exist_ok=True)
                        shutil.copy2(i, dst)

                def source(self):
                    self.output.info("SOURCES: {}!!".format(sorted(os.listdir("."))))
                    self.output.info("SOURCES_SUB: {}!!".format(sorted(os.listdir("sub"))))
            """)

        c = TestClient()
        c.save({"conanfile.py": conanfile,
                ".gitignore": "*.txt",
                "myfile.txt": "test",
                "myfile.other": "othertest",
                "sub/otherfile": "other"})
        c.init_git_repo()
        c.run("create .")
        assert "pkg/0.1: SOURCES: ['.gitignore', 'conanfile.py', 'myfile.other', 'sub']!!" in c.out
        assert "pkg/0.1: SOURCES_SUB: ['otherfile']!!" in c.out

    def test_git_included_subfolder(self):
        conanfile = textwrap.dedent("""
            import os
            import shutil
            from conan import ConanFile
            from conan.tools.scm import Git

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"

                def export_sources(self):
                    git = Git(self, "src")
                    included = git.included_files()
                    for i in included:
                        shutil.copy2(i, self.export_sources_folder)

                def source(self):
                    self.output.info("SOURCES: {}!!".format(sorted(os.listdir("."))))
            """)

        c = TestClient()
        c.save({"conanfile.py": conanfile,
                ".gitignore": "*.txt",
                "somefile": "some",
                "src/myfile.txt": "test",
                "src/myfile.other": "othertest"})
        c.init_git_repo()
        c.run("create .")
        assert "pkg/0.1: SOURCES: ['myfile.other']!!" in c.out


def test_capture_git_tag():
    """
    A local repo, without remote, will have commit, but no URL
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.scm import Git

        class Pkg(ConanFile):
            name = "pkg"

            def set_version(self):
                git = Git(self, self.recipe_folder)
                self.version = git.run("describe --tags")
        """)
    c.save({"conanfile.py": conanfile})
    c.init_git_repo()
    c.run_command("git tag 1.2")
    c.run("install .")
    assert "pkg/1.2" in c.out
    c.run("create .")
    assert "pkg/1.2" in c.out
    c.run("install --requires=pkg/1.2")
    assert "pkg/1.2" in c.out


@pytest.mark.tool("git")
class TestGitShallowTagClone:
    """
    When we do a shallow clone of a repo with a specific tag/branch, it doesn't
    clone any of the git history.  When we check to see if a commit is in the
    repo, we fallback to a git fetch if we can't verify the commit locally.
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.scm import Git

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def export(self):
                git = Git(self, self.recipe_folder)
                commit = git.get_commit()
                url = git.get_remote_url()
                self.output.info("URL: {}".format(url))
                self.output.info("COMMIT: {}".format(commit))
                in_remote = git.commit_in_remote(commit)
                self.output.info("COMMIT IN REMOTE: {}".format(in_remote))
                self.output.info("DIRTY: {}".format(git.is_dirty()))
        """)

    def test_find_tag_in_remote(self):
        """
        a shallow cloned repo won't have the new commit locally, but can fetch it.
        """
        folder = temp_folder()
        url, commit = create_local_git_repo(files={"conanfile.py": self.conanfile}, folder=folder)

        c = TestClient()
        # Create a tag
        with c.chdir(folder):
            c.run_command('git tag 1.0.0')

        # Do a shallow clone of our tag
        c.run_command('git clone --depth=1 --branch 1.0.0 "{}" myclone'.format(folder))
        with c.chdir("myclone"):
            c.run("export .")
            assert "pkg/0.1: COMMIT: {}".format(commit) in c.out
            assert "pkg/0.1: URL: {}".format(url) in c.out
            assert "pkg/0.1: COMMIT IN REMOTE: True" in c.out
            assert "pkg/0.1: DIRTY: False" in c.out

    def test_detect_commit_not_in_remote(self):
        """
        a shallow cloned repo won't have new commit in remote
        """
        folder = temp_folder()
        url, commit = create_local_git_repo(files={"conanfile.py": self.conanfile}, folder=folder)

        c = TestClient()
        # Create a tag
        with c.chdir(folder):
            c.run_command('git tag 1.0.0')

        # Do a shallow clone of our tag
        c.run_command('git clone --depth=1 --branch 1.0.0 "{}" myclone'.format(folder))
        with c.chdir("myclone"):
            c.save({"conanfile.py": self.conanfile + "\n# some coment!"})
            new_commit = git_add_changes_commit(c.current_folder)

            c.run("export .")
            assert "pkg/0.1: COMMIT: {}".format(new_commit) in c.out
            assert "pkg/0.1: URL: {}".format(url) in c.out
            assert "pkg/0.1: COMMIT IN REMOTE: False" in c.out
            assert "pkg/0.1: DIRTY: False" in c.out


@pytest.mark.tool("git")
class TestGitTreelessRemote:

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.scm import Git
        import os

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"

            def layout(self):
                self.folders.source = "source"

            def export(self):
                git = Git(self)
                git.clone(url="{url}", args=["--filter=tree:0"], target="target")
                git.folder = "target"
                cloned_url = git.run("remote -v")
                self.output.info("git remote: %s ===" % cloned_url)
                cloned_url = git.get_remote_url()
                self.output.info("get_remote_url(): %s ===" % cloned_url)
        """)

    def test_treeless_clone(self):
        """
        When cloning a git repository with the `--filter=tree:0` option,
        the Git.get_remote_url() should only the URL of the repository.

        Validate the issue https://github.com/conan-io/conan/issues/18415
        """
        repository = temp_folder(path_with_spaces=False)
        url, commit = create_local_git_repo(files={"README": "Lumen naturale ratum est"},
                                            folder=repository)

        client = TestClient()
        client.save({"conanfile.py": self.conanfile.format(url=url)})
        client.run("export .")
        # We expect [tree:0] for regular git remote command. Requires Git +2.43
        assert f"git remote: origin\t{url} (fetch) [tree:0]" in client.out
        # Then get_remote_url filters it to only the URL
        assert f"get_remote_url(): {url} ===" in client.out

    def test_treeless_clone_with_parenthesis(self):
        """
        Windows can use C:/Program Files (x86) path, which contains parenthesis.
        As the URL will have (fetch) or (push), get_remote_url() should be able to parse it.
        """
        repository = os.path.join(temp_folder(), "Program Files (x86)", "myrepo")
        os.makedirs(repository)
        url, commit = create_local_git_repo(files={"README": "Pacem in maribus."},
                                            folder=repository)

        client = TestClient()
        client.save({"conanfile.py": self.conanfile.format(url=url)})
        client.run("export .")
        assert f"get_remote_url(): {url} ===" in client.out



import pytest

from conan.tools.build import cross_building
from conan.test.utils.mocks import ConanFileMock


@pytest.mark.parametrize("cross_build", (True, False))
def test_using_cross_build_conf(cross_build):
    """
    Tests cross_building function is using the conf variable to force or not.

    Issue related: https://github.com/conan-io/conan/issues/15392
    """
    conanfile = ConanFileMock()
    conanfile.conf.define("tools.build.cross_building:cross_build", cross_build)
    assert cross_building(conanfile) == cross_build



import json
import os
import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient
from conan.test.utils.env import environment_update


class TestCustomCommandsErrors:

    def test_import_error_custom_command(self):
        mycommand = textwrap.dedent("""
            import this_doesnt_exist
            """)

        client = TestClient()
        command_file_path = os.path.join(client.cache_folder, 'extensions',
                                         'commands', 'cmd_mycommand.py')
        client.save({f"{command_file_path}": mycommand})
        # Call to any other command, it will fail loading the custom command
        client.run("list *")
        assert "ERROR: Error loading custom command 'cmd_mycommand.py': " \
               "No module named 'this_doesnt_exist'" in client.out
        # But it won't break the whole conan and you can still use the rest of it
        client.run("config home")
        assert client.cache_folder in client.out

    def test_import_error_custom_command_subfolder(self):
        """
        used to break, this is handled differently in conan
        """
        mycommand = textwrap.dedent("""
            import this_doesnt_exist
            """)

        client = TestClient()
        command_file_path = os.path.join(client.cache_folder, 'extensions',
                                         'commands', 'mycompany', 'cmd_mycommand.py')
        client.save({f"{command_file_path}": mycommand})
        # Call to any other command, it will fail loading the custom command,
        client.run("list *")
        assert "ERROR: Error loading custom command mycompany.cmd_mycommand" in client.out
        # But it won't break the whole conan and you can still use the rest of it
        client.run("config home")
        assert client.cache_folder in client.out

    def test_import_error_bad_name(self):
        mycommand = textwrap.dedent("""
            from conan.cli.command import conan_command, conan_subcommand
            @conan_command(group="custom commands")
            def mycommand(conan_api, parser, *args, **kwargs):
                \""" custom \"""
            @conan_subcommand()
            def mysubcmd(conan_api, parser, *args, **kwargs):
                \""" mysubcmd \"""
            """)

        c = TestClient()
        c.save_home({"extensions/commands/cmd_mycommand.py": mycommand})
        # Call to any other command, it will fail loading the custom command
        c.run("list *")
        assert "The name for the subcommand method should begin with the main command name" in c.out


class TestCustomCommands:
    def test_simple_custom_command(self):
        mycommand = textwrap.dedent("""
            import json
            import os

            from conan.cli.command import conan_command
            from conan.api.output import cli_out_write

            def output_mycommand_cli(info):
                cli_out_write(f"Conan cache folder is: {info.get('cache_folder')}")

            def output_mycommand_json(info):
                cli_out_write(json.dumps(info))

            @conan_command(group="custom commands",
                           formatters={"cli": output_mycommand_cli,
                                       "json": output_mycommand_json})
            def mycommand(conan_api, parser, *args, **kwargs):
                \"""
                this is my custom command, it will print the location of the cache folder
                \"""
                info = {"cache_folder": os.path.basename(conan_api.cache_folder)}
                return info
            """)

        client = TestClient()
        command_file_path = os.path.join(client.cache_folder, 'extensions',
                                         'commands', 'cmd_mycommand.py')
        client.save({f"{command_file_path}": mycommand})
        client.run("mycommand -f cli")
        foldername = os.path.basename(client.cache_folder)
        assert f'Conan cache folder is: {foldername}' in client.out
        client.run("mycommand -f json")
        assert f'{{"cache_folder": "{foldername}"}}' in client.out

    def test_command_layer(self):
        myhello = textwrap.dedent("""
            from conan.api.output import cli_out_write
            from conan.cli.command import conan_command

            @conan_command(group="custom commands")
            def hello(conan_api, parser, *args, **kwargs):
                '''
                My Hello doc
                '''
                cli_out_write("Hello {}!")
            """)
        mybye = textwrap.dedent("""
            from conan.api.output import cli_out_write
            from conan.cli.command import conan_command, conan_subcommand

            @conan_command(group="custom commands")
            def bye(conan_api, parser, *args, **kwargs):
                '''
                My Bye doc
                '''

            @conan_subcommand()
            def bye_say(conan_api, parser, *args, **kwargs):
                '''
                My bye say doc
                '''
                cli_out_write("Bye!")
            """)

        client = TestClient()
        layer_path = os.path.join(client.cache_folder, 'extensions', 'commands')
        client.save({os.path.join(layer_path, 'cmd_hello.py'): myhello.format("world"),
                     os.path.join(layer_path, "greet", 'cmd_hello.py'): myhello.format("moon"),
                     os.path.join(layer_path, "greet", 'cmd_bye.py'): mybye})
        # Test that the root "hello" without subfolder still works and no conflict
        client.run("hello")
        assert "Hello world!" in client.out
        client.run("greet:hello")
        assert "Hello moon!" in client.out
        client.run("greet:bye say")
        assert "Bye!" in client.out
        client.run("-h")
        assert "greet:bye" in client.out

        # Ensure the prog has the full command name
        client.run("hello -h")
        assert "conan hello" in client.out
        client.run("greet:bye -h")
        assert "conan greet:bye" in client.out
        client.run("greet:bye say -h")
        assert "conan greet:bye say" in client.out

    def test_custom_command_with_subcommands(self):
        complex_command = textwrap.dedent("""
            import json

            from conan.cli.command import conan_command, conan_subcommand
            from conan.api.output import cli_out_write

            def output_cli(info):
                cli_out_write(f"{info.get('argument1')}")

            def output_json(info):
                 cli_out_write(json.dumps(info))

            @conan_subcommand(formatters={"cli": output_cli, "json": output_json})
            def complex_sub1(conan_api, parser, subparser, *args):
                \"""
                sub1 subcommand
                \"""
                subparser.add_argument("argument1", help="This is argument number 1")
                args = parser.parse_args(*args)
                info = {"argument1": args.argument1}
                return info

            @conan_command()
            def complex(conan_api, parser, *args, **kwargs):
                \"""
                this is a command with subcommands
                \"""
            """)

        client = TestClient()
        command_file_path = os.path.join(client.cache_folder, 'extensions',
                                         'commands', 'cmd_complex.py')
        client.save({f"{command_file_path}": complex_command})
        client.run("complex sub1 myargument -f=cli")
        assert "myargument" in client.out
        client.run("complex sub1 myargument -f json")
        assert f'{{"argument1": "myargument"}}' in client.out

    def test_custom_command_with_subcommands_with_underscore(self):
        complex_command = textwrap.dedent("""
            import json

            from conan.cli.command import conan_command, conan_subcommand
            from conan.api.output import cli_out_write

            @conan_command()
            def command_with_underscores(conan_api, parser, *args, **kwargs):
                \"""
                this is a command with subcommands
                \"""

            @conan_subcommand()
            def command_with_underscores_subcommand_with_underscores_too(conan_api, parser, subparser, *args):
                \"""
                sub1 subcommand
                \"""
                subparser.add_argument("argument1", help="This is argument number 1")
                args = parser.parse_args(*args)
                cli_out_write(args.argument1)
            """)

        client = TestClient()
        command_file_path = os.path.join(client.cache_folder, 'extensions',
                                         'commands', 'cmd_command_with_underscores.py')
        client.save({f"{command_file_path}": complex_command})
        client.run("command-with-underscores subcommand-with-underscores-too myargument")
        assert "myargument" in client.out

    def test_overwrite_builtin_command(self):
        complex_command = textwrap.dedent("""
            import json

            from conan.cli.command import conan_command
            from conan.api.output import cli_out_write

            @conan_command()
            def install(conan_api, parser, *args, **kwargs):
                \"""
                this is a command with subcommands
                \"""
                cli_out_write("Hello world")
            """)

        client = TestClient()
        command_file_path = os.path.join(client.cache_folder, 'extensions',
                                         'commands', 'myteam', 'cmd_install.py')
        client.save({f"{command_file_path}": complex_command})
        command_file_path = os.path.join(client.cache_folder, 'extensions',
                                         'commands', 'cmd_install.py')
        client.save({f"{command_file_path}": complex_command})
        client.run("myteam:install")
        assert "Hello world" in client.out
        client.run("install")
        assert "Hello world" in client.out

    def test_custom_command_local_import(self):
        mycode = textwrap.dedent("""
            from conan.api.output import cli_out_write


            def write_output(folder):
                cli_out_write(f"Conan cache folder from cmd_mycode: {folder}")
        """)
        mycommand = textwrap.dedent("""
            import os

            from conan.cli.command import conan_command
            from mycode import write_output


            @conan_command(group="custom commands")
            def mycommand(conan_api, parser, *args, **kwargs):
                \"""
                this is my custom command, it will print the location of the cache folder
                \"""
                folder = os.path.basename(conan_api.cache_folder)
                write_output(folder)
            """)

        client = TestClient()
        mycommand_file_path = os.path.join(client.cache_folder, 'extensions',
                                           'commands', 'danimtb', 'cmd_mycommand.py')
        mycode_file_path = os.path.join(client.cache_folder, 'extensions',
                                        'commands', 'danimtb', 'mycode.py')
        client.save({
            mycode_file_path: mycode,
            mycommand_file_path: mycommand
        })
        client.run("danimtb:mycommand")
        foldername = os.path.basename(client.cache_folder)
        assert f'Conan cache folder from cmd_mycode: {foldername}' in client.out

    def test_custom_command_from_other_location(self):
        """
        Tests that setting developer env variable ``_CONAN_INTERNAL_CUSTOM_COMMANDS_PATH``
        will append that folder to the Conan custom command default location.
        """
        myhello = textwrap.dedent("""
            from conan.api.output import cli_out_write
            from conan.cli.command import conan_command

            @conan_command(group="custom commands")
            def hello(conan_api, parser, *args, **kwargs):
                '''
                My Hello doc
                '''
                cli_out_write("Hello {}!")
            """)

        client = TestClient()
        my_local_layer_path = temp_folder(path_with_spaces=False)
        layer_path = os.path.join(client.cache_folder, 'extensions', 'commands')
        client.save({os.path.join(layer_path, 'cmd_hello.py'): myhello.format("world")})
        client.save({"cmd_hello.py": myhello.format("Overridden")}, path=my_local_layer_path)
        with environment_update({"_CONAN_INTERNAL_CUSTOM_COMMANDS_PATH": my_local_layer_path}):
            client.run("hello")
            # Local commands have preference over Conan custom ones if they collide
            assert "Hello Overridden!" in client.out
        # Without the variable it only loads the default custom commands location
        client.run("hello")
        assert "Hello world!" in client.out


class TestCommandAPI:
    @pytest.mark.parametrize("argument", ['["list", "pkg*", "-c"]',
                                          '"list pkg* -c"'])
    def test_command_reuse_interface(self, argument):
        mycommand = textwrap.dedent(f"""
            import json
            from conan.cli.command import conan_command
            from conan.api.output import cli_out_write

            @conan_command(group="custom commands")
            def mycommand(conan_api, parser, *args, **kwargs):
                \""" mycommand help \"""
                result = conan_api.command.run({argument})
                cli_out_write(json.dumps(result["results"], indent=2))
            """)

        c = TestClient()
        command_file_path = os.path.join(c.cache_folder, 'extensions',
                                         'commands', 'cmd_mycommand.py')
        c.save({f"{command_file_path}": mycommand})
        c.run("mycommand", redirect_stdout="file.json")
        assert json.loads(c.load("file.json")) == {"Local Cache": {}}

    def test_command_reuse_other_custom(self):
        cmd1 = textwrap.dedent(f"""
            from conan.cli.command import conan_command
            from conan.api.output import cli_out_write

            @conan_command(group="custom commands")
            def mycmd1(conan_api, parser, *args, **kwargs):
                \"""mycommand help \"""
                # result = conan_api.command.run("")
                cli_out_write("MYCMD1!!!!!")
                conan_api.command.run("mycmd2")
            """)
        cmd2 = textwrap.dedent(f"""
            from conan.cli.command import conan_command
            from conan.api.output import cli_out_write

            @conan_command(group="custom commands")
            def mycmd2(conan_api, parser, *args, **kwargs):
                \"""mycommand help\"""
                cli_out_write("MYCMD2!!!!!")
            """)

        c = TestClient()
        cmds = os.path.join(c.cache_folder, 'extensions', 'commands')
        c.save({os.path.join(cmds, "cmd_mycmd1.py"): cmd1,
                os.path.join(cmds, "cmd_mycmd2.py"): cmd2})
        c.run("mycmd1")
        assert "MYCMD1!!!!!" in c.out
        assert "MYCMD2!!!!!" in c.out

    def test_command_verbosity_leak(self):
        mycommand = textwrap.dedent(f"""
                    import json
                    from conan.cli.command import conan_command
                    from conan.api.output import ConanOutput

                    @conan_command(group="custom commands")
                    def mycommand(conan_api, parser, *args, **kwargs):
                        \""" mycommand help \"""
                        parser.add_argument("foo", help="foo")
                        args = parser.parse_args(*args)

                        out = ConanOutput()
                        out.title("This is my first title")
                        conan_api.command.run("config home")
                        out.title("This is my second title")
                    """)

        c = TestClient()
        command_file_path = os.path.join(c.cache_folder, 'extensions',
                                         'commands', 'cmd_mycommand.py')
        c.save({f"{command_file_path}": mycommand})
        c.run("mycommand foo -vquiet")
        assert "This is my first title" not in c.out
        assert "This is my second title" not in c.out

    def test_command_reuse_interface_create(self):
        mycommand = textwrap.dedent("""
            import json
            from conan.cli.command import conan_command
            from conan.cli.formatters.graph import format_graph_json

            @conan_command(group="custom commands", formatters={"json": format_graph_json})
            def mycommand(conan_api, parser, *args, **kwargs):
                \""" mycommand help \"""
                result = conan_api.command.run(["create", ".", "--version=1.0.0"])
                return result
            """)

        c = TestClient()
        command_file_path = os.path.join(c.cache_folder, 'extensions',
                                         'commands', 'cmd_mycommand.py')
        c.save({f"{command_file_path}": mycommand,
                "conanfile.py": GenConanfile("mylib")})
        c.run("mycommand --format=json", redirect_stdout="file.json")
        create_output = json.loads(c.load("file.json"))
        assert create_output['graph']['nodes']['1']['label'] == "mylib/1.0.0"

    def test_subcommand_reuse_interface(self):
        mycommand = textwrap.dedent("""
            import json
            from conan.cli.command import conan_command
            from conan.api.output import cli_out_write

            @conan_command(group="custom commands")
            def mycommand(conan_api, parser, *args, **kwargs):
                \""" mycommand help \"""
                parser.add_argument("remote", help="remote")
                parser.add_argument("url", help="url")
                args = parser.parse_args(*args)
                conan_api.command.run(["remote", "add", args.remote, args.url])
                result = conan_api.command.run(["remote", "list"])
                result = {r.name: r.url for r in result}
                cli_out_write(json.dumps(result, indent=2))
            """)

        c = TestClient()
        command_file_path = os.path.join(c.cache_folder, 'extensions',
                                         'commands', 'cmd_mycommand.py')
        c.save({f"{command_file_path}": mycommand})
        c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        c.run("export .")
        c.run("mycommand myremote myurl")
        assert json.loads(c.stdout) == {"myremote": "myurl"}


class TestCommandsRemoteCaching:
    def test_remotes_cache(self):
        complex_command = textwrap.dedent("""
            import json
            from conan.cli.command import conan_command
            from conan.api.output import ConanOutput

            @conan_command()
            def mycache(conan_api, parser, *args, **kwargs):
                \""" this is a command with subcommands \"""

                remotes = conan_api.remotes.list()
                host = conan_api.profiles.get_profile(["default"])
                build = conan_api.profiles.get_profile(["default"])

                deps_graph = conan_api.graph.load_graph_requires(["pkg/0.1"], None, host, build,
                                                                 None, remotes, None, None)
                conan_api.graph.analyze_binaries(deps_graph, None, remotes=remotes)

                # SECOND RUN!!!
                # This will break if the caching is not working
                remotes[0].url = "broken"
                deps_graph = conan_api.graph.load_graph_requires(["pkg/0.1"], None, host, build,
                                                                 None, remotes, None, None)
                conan_api.graph.analyze_binaries(deps_graph, None, remotes=remotes)

                # Now invalidate the cache and see how it breaks
                try:
                    remotes[0].invalidate_cache()
                    deps_graph = conan_api.graph.load_graph_requires(["pkg/0.1"], None, host, build,
                                                                     None, remotes, None, None)
                    conan_api.graph.analyze_binaries(deps_graph, None, remotes=remotes)
                except Exception as e:
                    ConanOutput().warning(f"Cache invalidated, as expected: {e}")
            """)

        c = TestClient(default_server_user=True)
        c.save_home({"extensions/commands/cmd_mycache.py": complex_command})

        c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
                "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_require("dep/0.1")})
        c.run("create dep")
        c.run("create pkg")
        c.run("upload * -r=default -c")
        c.run("remove * -c")
        c.run("mycache")
        # Does not break unexpectedly, caching is working
        assert "WARN: Cache invalidated, as expected: Invalid URL 'broken" in c.out


def test_custom_command_settings_access():
    tc = TestClient()
    mycommand = textwrap.dedent("""
        from conan.cli.command import conan_command
        from conan.api.output import ConanOutput

        @conan_command(group="custom commands")
        def mycommand(conan_api, parser, *args, **kwargs):
            \"""
            test
            \"""
            settings = conan_api.config.settings_yml
            ConanOutput().info(f"settings.fields: {settings.fields}")
            ConanOutput().info(f"settings.possible_values(): {settings.possible_values()}")
            ConanOutput().info(f"settings.compiler: {settings.compiler}")
            ConanOutput().info(f"settings.compiler.possible_values(): {settings.compiler.possible_values()}")
        """)
    command_file_path = os.path.join('extensions', 'commands', 'cmd_mycommand.py')
    tc.save_home({f"{command_file_path}": mycommand})
    tc.run("mycommand")
    assert "settings.fields: ['arch', 'build_type', 'compiler', 'os']" in tc.out
    # No values here, but as it does not fail, the interface is working fine
    assert "settings.compiler: None" in tc.out



from conan.test.utils.mocks import ConanFileMock, MockSettings
from conan.errors import ConanException
from conan.tools.android import android_abi

from pytest import raises


def test_tools_android_abi():
    settings_linux = MockSettings({"os": "Linux", "arch": "foo"})

    for (arch, expected) in [
        ("armv5el", "armeabi"),
        ("armv5hf", "armeabi"),
        ("armv5", "armeabi"),
        ("armv6", "armeabi-v6"),
        ("armv7", "armeabi-v7a"),
        ("armv7hf", "armeabi-v7a"),
        ("armv8", "arm64-v8a"),
        ("x86", "x86"),
        ("x86_64", "x86_64"),
        ("mips", "mips"),
        ("mips_64", "mips_64"),
    ]:
        conanfile = ConanFileMock()
        settings_android = MockSettings({"os": "Android", "arch": arch})

        # 2 profiles
        ## native build
        conanfile.settings = settings_android
        conanfile.settings_host = settings_android
        conanfile.settings_build = settings_android

        assert android_abi(conanfile) == expected
        assert android_abi(conanfile, context="host") == expected
        assert android_abi(conanfile, context="build") == expected

        with raises(ConanException):
            assert android_abi(conanfile, context="target") == expected

        ## cross-build from Android to Linux (quite hypothetical)
        conanfile.settings = settings_linux
        conanfile.settings_host = settings_linux
        conanfile.settings_build = settings_android
        assert android_abi(conanfile) != expected
        assert android_abi(conanfile, context="host") != expected
        assert android_abi(conanfile, context="build") == expected

        with raises(ConanException):
            assert android_abi(conanfile, context="target")

        ## cross-build a recipe from Linux to Android:
        ### test android_abi in recipe itself
        conanfile.settings = settings_android
        conanfile.settings_host = settings_android
        conanfile.settings_build = settings_linux
        assert android_abi(conanfile) == expected
        assert android_abi(conanfile, context="host") == expected
        assert android_abi(conanfile, context="build") != expected
        with raises(ConanException):
            android_abi(conanfile, context="target")

        ### test android_abi in "compiler recipe" (ie a android-ndk recipe in tool_requires of recipe being cross-build)
        conanfile.settings = settings_linux
        conanfile.settings_host = settings_linux
        conanfile.settings_build = settings_linux
        conanfile.settings_target = settings_android
        assert android_abi(conanfile) != expected
        assert android_abi(conanfile, context="host") != expected
        assert android_abi(conanfile, context="build") != expected
        assert android_abi(conanfile, context="target") == expected



import os
import platform
import textwrap

import pytest

from conan.internal.api.profile.detect import detect_defaults_settings
from conan.test.utils.mocks import RedirectedTestOutput
from conan.test.utils.tools import TestClient, redirect_output
from conan.test.utils.env import environment_update
from conan.internal.util.files import save
from conan.internal.util.runners import detect_runner
from conan.tools.microsoft.visual import vcvars_command


class TestProfile:

    def test_list_empty(self):
        client = TestClient()
        client.run("profile list")
        assert "Profiles found in the cache:" in client.out

    def test_list(self):
        client = TestClient()
        profiles = ["default", "profile1", "profile2", "profile3",
                    "nested" + os.path.sep + "profile4",
                    "nested" + os.path.sep + "two" + os.path.sep + "profile5",
                    "nested" + os.path.sep + "profile6"]
        if platform.system() != "Windows":
            profiles.append("symlink_me" + os.path.sep + "profile7")
        for profile in profiles:
            save(os.path.join(client.paths.profiles_path, profile), "")

        if platform.system() != "Windows":
            os.symlink(os.path.join(client.paths.profiles_path, 'symlink_me'),
                       os.path.join(client.paths.profiles_path, 'link'))
            # profile7 will be shown twice because it is symlinked.
            profiles.append("link" + os.path.sep + "profile7")

        # Make sure local folder doesn't interact with profiles
        os.mkdir(os.path.join(client.current_folder, "profile3"))
        client.run("profile list")
        for p in profiles:
            assert p in client.out

        # Test profile list json file
        # FIXME: Json cannot be captured from tests?
        # client.run("profile list --format=json > profiles_list.json")
        # json_content = client.load("profiles_list.json")
        # json_obj = json.loads(json_content)
        # self.assertEqual(list, type(json_obj))
        # self.assertEqual(profiles, json_obj)

    def test_show(self):
        client = TestClient()
        profile1 = textwrap.dedent("""
            [settings]
            os=Windows
            [options]
            MyOption=32
            [conf]
            tools.build:jobs=20
            """)
        save(os.path.join(client.paths.profiles_path, "profile1"), profile1)

        client.run("profile show -pr=profile1")
        assert "[settings]\nos=Windows" in client.out
        assert "MyOption=32" in client.out
        # assert "CC=/path/tomy/gcc_build", client.out)
        # assert "CXX=/path/tomy/g++_build", client.out)
        # assert "package:VAR=value", client.out)
        assert "tools.build:jobs=20" in client.out

    def test_missing_subarguments(self):
        client = TestClient()
        client.run("profile", assert_error=True)
        assert "ERROR: Exiting with code: 2" in client.out


class TestDetectCompilers:
    def test_detect_default_compilers(self):
        platform_default_compilers = {
            "Linux": "gcc",
            "Darwin": "apple-clang",
            "Windows": "msvc"
        }

        result = detect_defaults_settings()
        # result is a list of tuples (name, value) so converting it to dict
        result = dict(result)
        platform_compiler = platform_default_compilers.get(platform.system(), None)
        if platform_compiler is not None:
            assert result.get("compiler", None) == platform_compiler

    @pytest.mark.tool("gcc")
    @pytest.mark.skipif(platform.system() != "Darwin", reason="only OSX test")
    def test_detect_default_in_mac_os_using_gcc_as_default(self):
        """
        Test if gcc in Mac OS X is using apple-clang as frontend
        """
        # See: https://github.com/conan-io/conan/issues/2231
        _, output = detect_runner("gcc --version")

        assert "clang" in output, "Apple gcc doesn't point to clang with gcc frontend anymore!"
        # Not test scenario gcc should display clang in output
        # see: https://stackoverflow.com/questions/19535422/os-x-10-9-gcc-links-to-clang

        output = RedirectedTestOutput()  # Initialize each command
        with redirect_output(output):
            with environment_update({"CC": "gcc"}):
                result = detect_defaults_settings()
        # result is a list of tuples (name, value) so converting it to dict
        result = dict(result)
        # No compiler should be detected
        assert result.get("compiler", None) is None
        assert "gcc detected as a frontend using apple-clang" in output

    def test_profile_new(self):
        c = TestClient()
        c.run("profile detect --name=./MyProfile2")
        profile = c.load("MyProfile2")
        assert "os=" in profile
        assert "compiler.runtime_type" not in profile  # Even in Windows

        c.run("profile detect --name=./MyProfile2", assert_error=True)
        assert "MyProfile2' already exists" in c.out

        c.run("profile detect --name=./MyProfile2 --force")  # will not raise error
        assert "build_type=Release" in c.load("MyProfile2")
        c.save({"MyProfile2": "potato"})
        c.run("profile detect --name=./MyProfile2 --exist-ok")  # wont raise, won't overwrite
        assert "Profile './MyProfile2' already exists, skipping detection" in c.out
        assert c.load("MyProfile2") == "potato"

    @pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows and msvc")
    def test_profile_new_msvc_vcvars(self):
        c = TestClient()

        # extract location of cl.exe from vcvars
        vcvars = vcvars_command(version="15", architecture="x64")
        ret = c.run_command(f"{vcvars} && where cl.exe")
        assert ret == 0
        cl_executable = c.out.splitlines()[-1]
        assert os.path.isfile(cl_executable)
        cl_location = os.path.dirname(cl_executable)

        # Try different variations, including full path and full path with quotes
        for var in ["cl", "cl.exe", cl_executable, f'"{cl_executable}"']:
            output = RedirectedTestOutput()
            with redirect_output(output):
                with environment_update({"CC": var, "PATH": cl_location}):
                    c.run("profile detect --name=./cl-profile --force")

            profile = c.load("cl-profile")
            assert "compiler=msvc" in profile
            assert "compiler.version=191" in profile



import platform
import textwrap

import pytest

from conan.test.utils.tools import TestClient


@pytest.mark.tool("cmake", "3.23")
def test_cmake_presets_with_user_presets_file():
    """ Test the integration of the generated one with a user root CMakePresets.json
    """
    c = TestClient()
    c.run("new cmake_exe -d name=foo -d version=1.0")
    conanfile = c.load("conanfile.py")
    conanfile = conanfile.replace("tc = CMakeToolchain(self)",
                                  "tc = CMakeToolchain(self)\n"
                                  "        tc.user_presets_path = 'ConanPresets.json'\n"
                                  "        tc.presets_prefix = 'conan'\n")
    cmake_presets = textwrap.dedent("""
        {
        "version": 4,
        "include": ["./ConanPresets.json"],
        "configurePresets": [
            {
                "name": "default",
                "displayName": "multi config",
                "inherits": "conan-default"
            },
            {
                "name": "release",
                "displayName": "release single config",
                "inherits": "conan-release"
            },
            {
                "name": "debug",
                "displayName": "debug single config",
                "inherits": "conan-debug"
            }
        ],
        "buildPresets": [
            {
                "name": "multi-release",
                "configurePreset": "default",
                "configuration": "Release",
                "inherits": "conan-release"
            },
            {
                "name": "multi-debug",
                "configurePreset": "default",
                "configuration": "Debug",
                "inherits": "conan-debug"
            },
            {
                "name": "release",
                "configurePreset": "release",
                "configuration": "Release",
                "inherits": "conan-release"
            },
            {
                "name": "debug",
                "configurePreset": "debug",
                "configuration": "Debug",
                "inherits": "conan-debug"
            }
        ]
        }""")
    c.save({"conanfile.py": conanfile,
            "CMakePresets.json": cmake_presets})

    c.run(f"install . ")
    c.run(f"install . -s build_type=Debug")

    if platform.system() != "Windows":
        c.run_command("cmake --preset debug")
        c.run_command("cmake --build --preset debug")
        c.run_command("./build/Debug/foo")
    else:
        c.run_command("cmake --preset default")
        c.run_command("cmake --build --preset multi-debug")
        c.run_command("build\\Debug\\foo")

    assert "Hello World Debug!" in c.out

    if platform.system() != "Windows":
        c.run_command("cmake --preset release")
        c.run_command("cmake --build --preset release")
        c.run_command("./build/Release/foo")
    else:
        c.run_command("cmake --build --preset multi-release")
        c.run_command("build\\Release\\foo")

    assert "Hello World Release!" in c.out



import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_build_requires_options_different():
    # copied from https://github.com/conan-io/conan/pull/9839
    # This is a test that crashed in 1.X, because of conflicting options
    client = TestClient()

    conanfile_openssl_1_1_1 = GenConanfile("openssl", "1.1.1")
    conanfile_openssl_3_0_0 = GenConanfile("openssl", "3.0.0") \
        .with_option("no_fips", [True, False]) \
        .with_default_option("no_fips", True)
    conanfile_cmake = GenConanfile("cmake", "0.1") \
        .with_requires("openssl/1.1.1")
    conanfile_consumer = GenConanfile("consumer", "0.1") \
        .with_build_requires("cmake/0.1") \
        .with_requires("openssl/3.0.0")

    client.save({"openssl_1_1_1.py": conanfile_openssl_1_1_1,
                 "openssl_3_0_0.py": conanfile_openssl_3_0_0,
                 "conanfile_cmake.py": conanfile_cmake,
                 "conanfile.py": conanfile_consumer})

    client.run("create openssl_1_1_1.py")
    client.run("create openssl_3_0_0.py")
    client.run("create conanfile_cmake.py")
    client.run("install conanfile.py")
    # This test used to crash, not crashing means ok
    assert "openssl/1.1.1" in client.out
    assert "openssl/3.0.0: Already installed!" in client.out
    assert "cmake/0.1: Already installed!" in client.out


def test_different_options_values_profile():
    """
    consumer -> protobuf (library)
        \\--(build)-> protobuf (protoc)
    protobuf by default is a static library (shared=False)
    The profile or CLI args can select for each one (library and protoc) the "shared" value
    """
    c = TestClient()
    protobuf = textwrap.dedent("""
        from conan import ConanFile
        class Proto(ConanFile):
            options = {"shared": [True, False]}
            default_options = {"shared": False}

            def package_info(self):
                self.output.info("MYOPTION: {}-{}".format(self.context, self.options.shared))
        """)

    c.save({"protobuf/conanfile.py": protobuf,
            "consumer/conanfile.py": GenConanfile().with_requires("protobuf/1.0")
           .with_build_requires("protobuf/1.0")})

    c.run("create protobuf --name=protobuf --version=1.0")
    c.run("create protobuf --name=protobuf --version=1.0 -o protobuf/*:shared=True")
    c.run("install consumer")
    assert "protobuf/1.0: MYOPTION: host-False" in c.out
    assert "protobuf/1.0: MYOPTION: build-False" in c.out
    # specifying it in the profile works
    c.run("install consumer -o protobuf/*:shared=True")
    assert "protobuf/1.0: MYOPTION: host-True" in c.out
    assert "protobuf/1.0: MYOPTION: build-False" in c.out
    c.run("install consumer -o protobuf/*:shared=True -o:b protobuf/*:shared=False")
    assert "protobuf/1.0: MYOPTION: host-True" in c.out
    assert "protobuf/1.0: MYOPTION: build-False" in c.out
    c.run("install consumer -o protobuf/*:shared=False -o:b protobuf/*:shared=True")
    assert "protobuf/1.0: MYOPTION: host-False" in c.out
    assert "protobuf/1.0: MYOPTION: build-True" in c.out
    c.run("install consumer -o:b protobuf/*:shared=True")
    assert "protobuf/1.0: MYOPTION: host-False" in c.out
    assert "protobuf/1.0: MYOPTION: build-True" in c.out


@pytest.mark.parametrize("scope", ["protobuf/*:", ""])
def test_different_options_values_recipe(scope):
    """
    consumer -> protobuf (library)
        \\--(build)-> protobuf (protoc)
    protobuf by default is a static library (shared=False)
    The "consumer" conanfile.py can use ``self.requires(...,options=)`` to define protobuf:shared
    """
    c = TestClient()
    protobuf = textwrap.dedent("""
        from conan import ConanFile
        class Proto(ConanFile):
            options = {"shared": [True, False]}
            default_options = {"shared": False}

            def package_info(self):
                self.output.info("MYOPTION: {}-{}".format(self.context, self.options.shared))
        """)
    consumer_recipe = textwrap.dedent("""
        from conan import ConanFile
        class Consumer(ConanFile):
            def requirements(self):
                self.requires("protobuf/1.0", options={{"{scope}shared": {host}}})
            def build_requirements(self):
                self.build_requires("protobuf/1.0", options={{"{scope}shared": {build}}})
        """)
    c.save({"conanfile.py": protobuf})

    c.run("create . --name=protobuf --version=1.0")
    c.run("create . --name=protobuf --version=1.0 -o protobuf/*:shared=True")

    for host, build in ((True, True), (True, False), (False, True), (False, False)):
        c.save({"conanfile.py": consumer_recipe.format(host=host, build=build, scope=scope)})
        c.run("install .")
        assert f"protobuf/1.0: MYOPTION: host-{host}" in c.out
        assert f"protobuf/1.0: MYOPTION: build-{build}" in c.out


def test_different_options_values_recipe_attributes():
    """
    consumer -> protobuf (library)
        \\--(build)-> protobuf (protoc)
    protobuf by default is a static library (shared=False)
    The "consumer" conanfile.py can use ``default_options`` to define protobuf:shared
    """
    c = TestClient()
    protobuf = textwrap.dedent("""
        from conan import ConanFile
        class Proto(ConanFile):
            options = {"shared": [True, False]}
            default_options = {"shared": False}

            def package_info(self):
                self.output.info("MYOPTION: {}-{}".format(self.context, self.options.shared))
        """)
    c.save({"conanfile.py": protobuf})
    c.run("create . --name=protobuf --version=1.0")
    c.run("create . --name=protobuf --version=1.0 -o protobuf/*:shared=True")

    consumer_recipe = textwrap.dedent("""
        from conan import ConanFile
        class Consumer(ConanFile):
            default_options = {{"protobuf/*:shared": {host}}}
            default_build_options = {{"protobuf/*:shared": {build}}}
            def requirements(self):
                self.requires("protobuf/1.0")
            def build_requirements(self):
                self.build_requires("protobuf/1.0")
        """)

    for host, build in ((True, True), (True, False), (False, True), (False, False)):
        c.save({"conanfile.py": consumer_recipe.format(host=host, build=build)})
        c.run("install .")
        assert f"protobuf/1.0: MYOPTION: host-{host}" in c.out
        assert f"protobuf/1.0: MYOPTION: build-{build}" in c.out


def test_different_options_values_recipe_priority():
    """
    consumer ---> mypkg ---> protobuf (library)
                  \\--(build)-> protobuf (protoc)
    protobuf by default is a static library (shared=1)
    "consumer" defines a protobuf:shared=3 value, that must be respected for HOST context
    But build context, it is assigned by "mypkg", and build-require is private
    """
    c = TestClient()
    protobuf = textwrap.dedent("""
        from conan import ConanFile
        class Proto(ConanFile):
            options = {"shared": [1, 2, 3]}
            default_options = {"shared": 1}

            def package_id(self):
                self.output.info("MYOPTION: {}-{}".format(self.context, self.info.options.shared))
        """)
    my_pkg = textwrap.dedent("""
        from conan import ConanFile
        class Consumer(ConanFile):
            def requirements(self):
                self.requires("protobuf/1.0", options={"shared": 2})
            def build_requirements(self):
                self.build_requires("protobuf/1.0", options={"shared": 2})
        """)
    c.save({"protobuf/conanfile.py": protobuf,
            "mypkg/conanfile.py": my_pkg,
            "consumer/conanfile.py": GenConanfile().with_requires("mypkg/1.0")
           .with_default_option("protobuf/*:shared", 3)})

    c.run("create protobuf --name=protobuf --version=1.0 -o protobuf/*:shared=2")
    c.run("create protobuf --name=protobuf --version=1.0 -o protobuf/*:shared=3")
    c.run("create mypkg --name=mypkg --version=1.0")

    c.run("install consumer")
    assert f"protobuf/1.0: MYOPTION: host-3" in c.out
    assert f"protobuf/1.0: MYOPTION: build-2" in c.out



import os
import textwrap

from conan.test.utils.tools import TestClient
from conan.internal.util.files import save


def test_profile_local_folder_priority_cache():
    """ includes or args without "./" will resolve to the cache first
    """
    c = TestClient()
    c.save({"profiles/default": f"include(otherprofile)",
            "profiles/otherprofile": "[settings]\nos=AIX",
            "conanfile.txt": ""})
    save(os.path.join(c.paths.profiles_path, "otherprofile"), "[settings]\nos=FreeBSD")

    # Must use local path, otherwise look for it in the cache
    c.run("install . -pr=./profiles/default")
    assert "os=FreeBSD" in c.out


def test_profile_local_folder_priority_relative():
    """ The local include(./profile) must have priority over a file with same name in cache
    """
    c = TestClient()
    c.save({"profiles/default": f"include(./otherprofile)",
            "profiles/otherprofile": "[settings]\nos=AIX",
            "conanfile.txt": ""})
    save(os.path.join(c.paths.profiles_path, "otherprofile"), "[settings]\nos=FreeBSD")

    # Must use local path, otherwise look for it in the cache
    c.run("install . -pr=./profiles/default")
    assert "os=AIX" in c.out


def test_profile_cache_folder_priority():
    """ The cache include(./profile) must have priority over a file with same name in local
    """
    c = TestClient()
    c.save({"otherprofile": "[settings]\nos=FreeBSD",
            "conanfile.txt": ""})
    save(os.path.join(c.paths.profiles_path, "default"), "include(./otherprofile)")
    save(os.path.join(c.paths.profiles_path, "otherprofile"), "[settings]\nos=AIX")

    c.run("install . -pr=default")
    assert "os=AIX" in c.out


def test_profile_cli_priority():
    c = TestClient()
    profile1 = textwrap.dedent("""\
        [settings]
        os=AIX
        [conf]
        user.myconf:myvalue1=1
        user.myconf:myvalue2=[2]
        user.myconf:myvalue3={"3": "4", "a": "b"}
        user.myconf:myvalue4={"1": "2"}
        user.myconf:myvalue5={"6": "7"}
        """)
    profile2 = textwrap.dedent("""\
        [settings]
        os=FreeBSD
        [conf]
        user.myconf:myvalue1=2
        user.myconf:myvalue2+=4
        user.myconf:myvalue3*={"3": "5"}
        user.myconf:myvalue5={"6": "7"}
        """)
    c.save({"profile1": profile1,
            "profile2": profile2})
    c.run("profile show -pr=./profile1 -pr=./profile2")
    assert "os=FreeBSD" in c.out
    assert "user.myconf:myvalue1=2" in c.out
    assert "user.myconf:myvalue2=[2, 4]" in c.out
    assert "user.myconf:myvalue3={'3': '5', 'a': 'b'}" in c.out
    assert "user.myconf:myvalue4={'1': '2'}" in c.out
    assert "user.myconf:myvalue5={'6': '7'}" in c.out


def test_profiles_patterns_include():
    # https://github.com/conan-io/conan/issues/16718
    c = TestClient()
    msvc = textwrap.dedent("""
        [settings]
        compiler=msvc
        compiler.cppstd=14
        compiler.version=193
        os=Windows

        test*/*:compiler.cppstd=14
        """)
    clang = textwrap.dedent("""
        include(./msvc)
        [settings]
        test*/*:compiler=clang
        test*/*:compiler.cppstd=17
        test*/*:compiler.runtime_version=v144
        test*/*:compiler.version=18
        """)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "test_pkg"
            version = "0.1"
            settings = "os", "compiler"
            def generate(self):
                self.output.info(f"MyCompiler={self.settings.compiler}!!!")
                self.output.info(f"MyCompilerVersion={self.settings.compiler.version}!!!")
                self.output.info(f"MyCompilerCpp={self.settings.compiler.cppstd}!!!")
            """)
    c.save({"conanfile.py": conanfile,
            "msvc": msvc,
            "clang": clang})
    c.run("install . -pr=clang")
    assert "conanfile.py (test_pkg/0.1): MyCompiler=clang!!!" in c.out
    assert "conanfile.py (test_pkg/0.1): MyCompilerVersion=18!!!" in c.out
    assert "conanfile.py (test_pkg/0.1): MyCompilerCpp=17!!!" in c.out



import pytest

from conan.tools.build import default_cppstd, default_cstd
from conan.tools.build.flags import architecture_flag, build_type_flags, threads_flags
from conan.test.utils.mocks import MockSettings, ConanFileMock


class TestCompilerFlags:

    @pytest.mark.parametrize("compiler,arch,the_os,flag",
                             [("gcc", "x86", None, "-m32"),
                              ("clang", "x86", None, "-m32"),
                              ("sun-cc", "x86", None, "-m32"),
                              ("gcc", "x86_64", None, "-m64"),
                              ("clang", "x86_64", None, "-m64"),
                              ("sun-cc", "x86_64", None, "-m64"),
                              ("sun-cc", "sparc", None, "-m32"),
                              ("sun-cc", "sparcv9", None, "-m64"),
                              ("gcc", "armv7", None, ""),
                              ("clang", "armv7", None, ""),
                              ("sun-cc", "armv7", None, ""),
                              ("gcc", "s390", None, "-m31"),
                              ("clang", "s390", None, "-m31"),
                              ("sun-cc", "s390", None, "-m31"),
                              ("gcc", "s390x", None, "-m64"),
                              ("clang", "s390x", None, "-m64"),
                              ("sun-cc", "s390x", None, "-m64"),
                              ("msvc", "x86", None, ""),
                              ("msvc", "x86_64", None, ""),
                              ("gcc", "ppc32", "AIX", "-maix32"),
                              ("gcc", "ppc64", "AIX", "-maix64"),
                              ])
    def test_arch_flag(self, compiler, arch, the_os, flag):
        settings = MockSettings({"compiler": compiler,
                                 "arch": arch,
                                 "os": the_os})
        conanfile = ConanFileMock()
        conanfile.settings = settings
        assert architecture_flag(conanfile) == flag

    @pytest.mark.parametrize("compiler,threads,flag",
                             [("clang", None, []),
                              ("emcc", None, []),
                              ("emcc", "posix", ["-pthread"]),
                              ("emcc", "wasm_workers", ["-sWASM_WORKERS=1"]),
                              ])
    def test_threads_flag(self, compiler, threads, flag):
        settings = MockSettings({"compiler": compiler,
                                 "compiler.threads": threads})
        conanfile = ConanFileMock()
        conanfile.settings = settings
        assert threads_flags(conanfile) == flag

    @pytest.mark.parametrize("compiler,arch,the_os,flag",
                             [("clang", "x86", "Windows", ""),
                              ("clang", "x86_64", "Windows", "")
                             ])
    def test_arch_flag_clangcl(self,  compiler, arch, the_os, flag):
        settings = MockSettings({"compiler": compiler,
                                 "arch": arch,
                                 "os": the_os})
        conanfile = ConanFileMock()
        conanfile.conf.define("tools.build:compiler_executables", {"c": "clang-cl"})
        conanfile.settings = settings
        assert architecture_flag(conanfile) == flag

    def test_catalyst(self):
        settings = MockSettings({"compiler": "apple-clang",
                                 "arch": "x86_64",
                                 "os": "Macos",
                                 "os.subsystem": "catalyst",
                                 "os.subsystem.ios_version": "13.1"})
        conanfile = ConanFileMock()
        conanfile.settings = settings
        assert architecture_flag(conanfile) == "--target=x86_64-apple-ios13.1-macabi"

        settings = MockSettings({"compiler": "apple-clang",
                                 "arch": "armv8",
                                 "os": "Macos",
                                 "os.subsystem": "catalyst",
                                 "os.subsystem.ios_version": "13.1"})
        conanfile = ConanFileMock()
        conanfile.settings = settings
        assert architecture_flag(conanfile) == "--target=arm64-apple-ios13.1-macabi"

    @pytest.mark.parametrize("os_,arch,flag",
                             [("Linux", "x86", "-m32"),
                              ("Linux", "x86_64", "-m64"),
                              ("Windows", "x86", "/Qm32"),
                              ("Windows", "x86_64", "/Qm64"),
                              ])
    def test_arch_flag_intel(self, os_, arch, flag):
        settings = MockSettings({"compiler": "intel-cc",
                                 "os": os_,
                                 "arch": arch})
        conanfile = ConanFileMock()
        conanfile.settings = settings
        assert architecture_flag(conanfile) == flag

    @pytest.mark.parametrize("arch,flag",
                             [("e2k-v2", "-march=elbrus-v2"),
                              ("e2k-v3", "-march=elbrus-v3"),
                              ("e2k-v4", "-march=elbrus-v4"),
                              ("e2k-v5", "-march=elbrus-v5"),
                              ("e2k-v6", "-march=elbrus-v6"),
                              ("e2k-v7", "-march=elbrus-v7"),
                              ])
    def test_arch_flag_mcst_lcc(self, arch, flag):
        conanfile = ConanFileMock()
        settings = MockSettings({"compiler": "mcst-lcc",
                                 "arch": arch})
        conanfile.settings = settings
        assert architecture_flag(conanfile) == flag

    @pytest.mark.parametrize("compiler,build_type,vs_toolset,flags",
                             [("msvc", "Debug", None, "-Zi -Ob0 -Od"),
                              ("msvc", "Release", None, "-O2 -Ob2"),
                              ("msvc", "RelWithDebInfo", None, "-Zi -O2 -Ob1"),
                              ("msvc", "MinSizeRel", None, "-O1 -Ob1"),
                              ("msvc", "Debug", "v140_clang_c2",
                               "-gline-tables-only -fno-inline -O0"),
                              ("msvc", "Release", "v140_clang_c2", "-O2"),
                              ("msvc", "RelWithDebInfo", "v140_clang_c2",
                               "-gline-tables-only -O2 -fno-inline"),
                              ("msvc", "MinSizeRel", "v140_clang_c2", ""),
                              ("gcc", "Debug", None, "-g"),
                              ("gcc", "Release", None, "-O3"),
                              ("gcc", "RelWithDebInfo", None, "-O2 -g"),
                              ("gcc", "MinSizeRel", None, "-Os"),
                              ("clang", "Debug", None, "-g"),
                              ("clang", "Release", None, "-O3"),
                              ("clang", "RelWithDebInfo", None, "-O2 -g"),
                              ("clang", "MinSizeRel", None, "-Os"),
                              ("apple-clang", "Debug", None, "-g"),
                              ("apple-clang", "Release", None, "-O3"),
                              ("apple-clang", "RelWithDebInfo", None, "-O2 -g"),
                              ("apple-clang", "MinSizeRel", None, "-Os"),
                              ("sun-cc", "Debug", None, "-g"),
                              ("sun-cc", "Release", None, "-xO3"),
                              ("sun-cc", "RelWithDebInfo", None, "-xO2 -g"),
                              ("sun-cc", "MinSizeRel", None, "-xO2 -xspace"),
                              ])
    def test_build_type_flags(self, compiler, build_type, vs_toolset, flags):
        settings = MockSettings({"compiler": compiler,
                                 "build_type": build_type,
                                 "compiler.toolset": vs_toolset})
        conanfile = ConanFileMock()
        conanfile.settings = settings
        assert ' '.join(build_type_flags(conanfile)) == flags

    @pytest.mark.parametrize("compiler,build_type,flags",
                             [("clang", "Debug", "-Zi -Ob0 -Od"),
                              ("clang", "Release", "-O2 -Ob2"),
                              ("clang", "RelWithDebInfo", "-Zi -O2 -Ob1"),
                              ("clang", "MinSizeRel", "-O1 -Ob1"),
                              ])
    def test_build_type_flags_clangcl(self, compiler, build_type, flags):
        settings = MockSettings({"compiler": compiler,
                                 "build_type": build_type})
        conanfile = ConanFileMock()
        conanfile.conf.define("tools.build:compiler_executables", {"c": "clang-cl"})
        conanfile.settings = settings
        assert ' '.join(build_type_flags(conanfile)) == flags

    @pytest.mark.parametrize("compiler,version,flags",
                             [("clang", "8", "gnu14"),
                              ("gcc", "15", "gnu17")
                              ])
    def test_default_cppstd(self, compiler, version, flags):
        settings = MockSettings({"compiler": compiler,
                                 "compiler.version": version})
        conanfile = ConanFileMock()
        conanfile.settings = settings
        assert default_cppstd(conanfile) == flags

    @pytest.mark.parametrize("compiler,version,flags",
                             [("clang", "8", "gnu11"),
                              ("gcc", "15", "gnu23")
                              ])
    def test_default_cstd(self, compiler, version, flags):
        settings = MockSettings({"compiler": compiler,
                                 "compiler.version": version})
        conanfile = ConanFileMock()
        conanfile.settings = settings
        assert default_cstd(conanfile) == flags



import os
import platform
import textwrap

from conan.test.utils.tools import TestClient


def test_multi_cmake():
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMakeToolchain, CMake, cmake_layout, CMakeDeps
        import os


        class multiRecipe(ConanFile):
            settings = "os", "compiler", "build_type", "arch"

            exports_sources = "cmake_one/CMakeLists.txt", "cmake_two/CMakeLists.txt", "src_one/*", "src_two/*"

            def layout(self):
                cmake_layout(self)

            def generate(self):
                deps = CMakeDeps(self)
                deps.generate()
                tc = CMakeToolchain(self)
                tc.generate()

            def build_one(self):
                cmake = CMake(self)
                cmake.configure(build_script_folder="cmake_one", subfolder="one")
                cmake.build(subfolder="one")
                # cmake.install(subfolder="one")

            def build_two(self):
                cmake = CMake(self)
                # CMAKE_PREFIX_PATH
                cmake.configure(build_script_folder="cmake_two", subfolder="two")
                cmake.build(subfolder="two")

            def build(self):
                self.build_one()
                self.build_two()

            def package(self):
                cmake = CMake(self)
                cmake.install(subfolder="two")

            def package_info(self):
                self.cpp_info.libs = ["hello_two"]
                self.cpp_info.includedirs = ['two/include']
                self.cpp_info.libdirs = ['two/lib']
        """)

    hello_cpp = textwrap.dedent("""
        #include <iostream>
        #include "hello_{name}.h"

        void hello_{name}() {{
            std::cout << "Hello, World {name}!" << std::endl;
        }}
        """)

    hello_h = textwrap.dedent("""
        #ifndef HELLO_{name}_H
        #define HELLO_{name}_H

        void hello_{name}();

        #endif
        """)

    cmakelist = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(hello_{name} LANGUAGES CXX)

        add_library(hello_{name} ../src_{name}/hello_{name}.cpp)
        target_include_directories(hello_{name} PUBLIC ../src_{name})

        set_target_properties(hello_{name} PROPERTIES PUBLIC_HEADER "../src_{name}/hello_{name}.h")
        install(TARGETS hello_{name})
        """)

    client = TestClient()
    client.save({"conanfile.py": conanfile,
                 "cmake_one/CMakeLists.txt": cmakelist.format(name="one"),
                 "cmake_two/CMakeLists.txt": cmakelist.format(name="two"),
                 "src_one/hello_one.h": hello_h.format(name="one"),
                 "src_one/hello_one.cpp": hello_cpp.format(name="one"),
                 "src_two/hello_two.h": hello_h.format(name="two"),
                 "src_two/hello_two.cpp": hello_cpp.format(name="two")})

    client.run("create . --name=multi --version=0.1")
    file_ext = '.a' if platform.system() != "Windows" else '.lib'
    lib_prefix = 'lib' if platform.system() != "Windows" else ''

    assert "multi/0.1: package(): Packaged 1 '.h' file: hello_two.h" in client.out
    assert f"multi/0.1: package(): Packaged 1 '{file_ext}' file: {lib_prefix}hello_two{file_ext}" in client.out
    package_folder = client.created_layout().package()
    assert not os.path.exists(os.path.join(package_folder, "one", "include", "hello_one.h"))
    assert not os.path.exists(os.path.join(package_folder, "one", "lib", f"{lib_prefix}hello_one{file_ext}"))
    assert os.path.exists(os.path.join(package_folder, "two", "include", "hello_two.h"))
    assert os.path.exists(os.path.join(package_folder, "two", "lib", f"{lib_prefix}hello_two{file_ext}"))



import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient

conanfile = """
from conan import ConanFile

class ExceptionsTest(ConanFile):
    name = "exceptions"
    version = "0.1"

    def {method}(self):
        {method_contents}

    def _aux_method(self):
        raise Exception('Oh! an error!')
"""


@pytest.mark.parametrize("direct", [True, False])
@pytest.mark.parametrize("method",
                         ["source", "build", "package", "package_info", "configure", "build_id",
                          "package_id", "requirements", "config_options", "layout", "generate",
                          "export", "export_sources", "build_requirements", "init"])
def test_all_methods(direct, method):
    client = TestClient()
    if direct:
        throw = "raise Exception('Oh! an error!')"
    else:
        throw = "self._aux_method()"

    client.save({"conanfile.py": conanfile.format(method=method, method_contents=throw)})
    client.run("create . ", assert_error=True)
    assert "Error in %s() method, line 9" % method in client.out
    assert "Oh! an error!" in client.out
    if not direct:
        assert "while calling '_aux_method', line 12" in client.out


def test_complete_traceback_debug():
    """
    in debug level (-vv), the trace is shown (this is for recipe methods exceptions)
    """
    client = TestClient()
    throw = "self._aux_method()"
    client.save({"conanfile.py": conanfile.format(method="source", method_contents=throw)})
    client.run("create . -vv", assert_error=True)
    assert "Exception: Oh! an error!" in client.out
    assert "ERROR: Traceback (most recent call last):" in client.out


def test_complete_traceback_trace():
    """
    in debug level (-vvv), the full trace is shown for ConanExceptions
    """
    client = TestClient()
    client.run("install --requires=pkg/1.0 -vvv", assert_error=True)
    assert "Traceback (most recent call last)" in client.out


def test_notracebacks_cmakedeps():
    """
    CMakeDeps prints traceback
    """
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_generator("CMakeDeps")})
    c.run("install .", assert_error=True)
    assert "Traceback" not in c.out
    assert "ERROR: Error in generator 'CMakeDeps'" in c.out



import json
import textwrap

from conan.test.utils.tools import GenConanfile, TestClient


def test_inject_generators_conf():
    tc = TestClient(light=True)
    tc.save({"tool/conanfile.py": textwrap.dedent("""
        from conan import ConanFile

        class MyGenerator:
            def __init__(self, conanfile):
                self.conanfile = conanfile

            def generate(self):
                self.conanfile.output.info("MyGenerator generated")

        class ToolConan(ConanFile):
            name = "tool"
            version = "0.1"

            def package_info(self):
                self.generator_info = ["CMakeToolchain", MyGenerator]
        """)})

    tc.save({"consumer/conanfile.py": GenConanfile("consumer", "0.1")
                .with_tool_requires("tool/0.1")
                .with_class_attribute("generators = 'VirtualBuildEnv', 'VirtualRunEnv'")})

    tc.run("create tool")
    tc.run("create consumer")
    assert "WARN: experimental: Tool-require tool/0.1 adding generators: " \
           "['CMakeToolchain', 'MyGenerator']" in tc.out
    assert "Generator 'CMakeToolchain' calling 'generate()'" in tc.out
    assert "Generator 'MyGenerator' calling 'generate()'" in tc.out
    assert "Generator 'VirtualBuildEnv' calling 'generate()'" in tc.out
    assert "Generator 'VirtualRunEnv' calling 'generate()'" in tc.out
    assert "CMakeToolchain generated: conan_toolchain.cmake" in tc.out
    assert "MyGenerator generated" in tc.out


def test_inject_generators_error():
    c = TestClient(light=True)
    c.save({"conanfile.py": textwrap.dedent("""
        from conan import ConanFile

        class ToolConan(ConanFile):
            name = "tool"
            version = "0.1"

            def package_info(self):
                self.generator_info = "CMakeToolchain"
        """)})
    c.run("create .")
    c.run("install --tool-requires=tool/0.1", assert_error=True)
    assert "ERROR: tool/0.1 'generator_info' must be a list" in c.out


def test_inject_vars():
    tc = TestClient(light=True)
    tc.save({
        "tool/envars_generator1.sh": textwrap.dedent("""
        export VAR1="value1"
        export PATH="$PATH:/additional/path"
        """),
        "tool/envars_generator2.sh": textwrap.dedent("""
        export VAR2="value2"
        """),
        "tool/conanfile.py": textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.env import create_env_script, register_env_script
        import os

        class GeneratorSet:
            def __init__(self, conanfile):
                self.conanfile = conanfile

            def generate(self):
                content = 'call envars_generator.sh'
                name = f"conantest.sh"
                create_env_script(self.conanfile, content, name)
                register_env_script(self.conanfile, "tool/envars_generator2.sh")

        class ToolConan(ConanFile):
            name = "tool"
            version = "0.1"

            def package_info(self):
                self.generator_info = [GeneratorSet]
        """)})

    tc.run("create tool")
    tc.run(f"install --tool-requires=tool/0.1")

    content = tc.load("conanbuild.sh")
    assert "conantest.sh" in content
    assert "envars_generator2.sh" in content


def test_json_serialization_error():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class MyGenerator:
            def __init__(self, conanfile):
                self._conanfile = conanfile

            def generate(self):
                pass

        class MyToolReq(ConanFile):
            name = "mygenerator-tool"
            version = "1.0"

            def package_info(self):
                self.generator_info = [MyGenerator]
        """)

    c.save({"tool/conanfile.py": conanfile})
    c.run("create tool")
    c.run("install --tool-requires=mygenerator-tool/1.0 --format=json")
    graph = json.loads(c.stdout)
    assert graph["graph"]["nodes"]["0"]["generators"] == ["MyGenerator"]



from collections import OrderedDict

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, TestServer


class TestVersionRangesUpdatingTest:

    def test_update_remote(self):
        # https://github.com/conan-io/conan/issues/5333
        client = TestClient(light=True, default_server_user=True)
        client.save({"conanfile.py": GenConanfile("boost")})
        client.run("create . --version=1.69.0")
        client.run("create . --version=1.70.0")
        client.run("upload * -r=default --confirm")
        client.run("remove * -c")

        client.save({"conanfile.txt": "[requires]\nboost/[*]"}, clean_first=True)
        client.run("install .")
        assert "boost/1.70" in client.out
        assert "boost/1.69" not in client.out

        client.run("install .")
        assert "boost/1.70" in client.out
        assert "boost/1.69" not in client.out

        client.run("install . --update")
        assert "boost/1.70" in client.out
        assert "boost/1.69" not in client.out

    def test_update(self):
        client = TestClient(light=True, default_server_user=True)

        client.save({"pkg/conanfile.py": GenConanfile("pkg"),
                     "app/conanfile.py": GenConanfile().with_requirement("pkg/[~1]")})
        client.run("create pkg --version=1.1")
        client.run("create pkg --version=1.2")
        client.run("upload * -r=default --confirm")
        client.run("remove pkg/1.2* -c")

        client.run("install app")
        # Resolves to local package
        assert "pkg/1.1" in client.out
        assert "pkg/1.2" not in client.out

        client.run("install app --update")
        # Resolves to remote package
        assert "pkg/1.1" not in client.out
        assert "pkg/1.2" in client.out

        # newer in cache that in remotes and updating, should resolve the cache one
        client.run("create pkg --version=1.3")
        client.run("install app --update")
        assert "pkg/1.2" not in client.out
        assert "pkg/1.3" in client.out
        client.run("remove pkg/1.3* -c")

        # removes remote
        client.run("remove pkg* -r=default -c")
        # Resolves to local package
        client.run("install app")
        assert "pkg/1.1" not in client.out
        assert "pkg/1.2" in client.out

        client.run("install app --update")
        assert "pkg/1.1" not in client.out
        assert "pkg/1.2" in client.out


class TestVersionRangesMultiRemote:

    def test_multi_remote(self):
        servers = OrderedDict()
        servers["default"] = TestServer()
        servers["other"] = TestServer()
        client = TestClient(light=True, servers=servers, inputs=2*["admin", "password"])
        client.save({"hello0/conanfile.py": GenConanfile("hello0"),
                     "hello1/conanfile.py": GenConanfile("hello1").with_requires("hello0/[*]")})
        client.run("export hello0 --version=0.1")
        client.run("export hello0 --version=0.2")
        client.run("upload * -r=default -c")
        client.run("export hello0 --version=0.3")
        client.run("upload hello0/0.3 -r=other -c")
        client.run('remove "hello0/*" -c')

        client.run("install hello1 --build missing -r=default")
        assert "hello0/0.2" in client.out
        assert "hello0/0.3" not in client.out
        client.run("remove hello0/* -c")
        client.run("install hello1 --build missing -r=other")
        assert "hello0/0.2" not in client.out
        assert "hello0/0.3" in client.out



import tempfile
from collections import namedtuple

import pytest

from conan.internal.rest.file_uploader import FileUploader
from conan.internal.errors import InternalErrorException, AuthenticationException, ForbiddenException
from conan.internal.util.files import save


class _ConfigMock:
    def get(self, name, default=None, check_type=None):
        return default


class MockRequester:
    retry = 0
    retry_wait = 0

    def __init__(self, response):
        self._response = response

    def put(self, *args, **kwargs):
        return namedtuple("response", "status_code content")(self._response, "tururu")


class TestUploaderUnit:
    @pytest.fixture(autouse=True)
    def setup(self):
        _, self.f = tempfile.mkstemp()
        save(self.f, "some contents")

    def test_401_raises_unauthoirzed_exception(self):
        uploader = FileUploader(MockRequester(401), verify=False, config=_ConfigMock())
        with pytest.raises(AuthenticationException, match="tururu"):
            uploader.upload("fake_url", self.f)

    def test_403_raises_unauthoirzed_exception_if_no_token(self):
        auth = namedtuple("auth", "bearer")(None)
        uploader = FileUploader(MockRequester(403), verify=False, config=_ConfigMock())
        with pytest.raises(AuthenticationException, match="tururu"):
            uploader.upload("fake_url", self.f, auth=auth)

    def test_403_raises_unauthorized_exception_if_no_auth(self):
        uploader = FileUploader(MockRequester(403), verify=False, config=_ConfigMock())
        with pytest.raises(AuthenticationException, match="tururu"):
            uploader.upload("fake_url", self.f)

    def test_403_raises_forbidden_exception_if_token(self):
        auth = namedtuple("auth", "bearer")("SOMETOKEN")
        uploader = FileUploader(MockRequester(403), verify=False, config=_ConfigMock())
        with pytest.raises(ForbiddenException, match="tururu"):
            uploader.upload("fake_url", self.f, auth=auth)

    def test_500_raises_internal_error(self):
        uploader = FileUploader(MockRequester(500), verify=False, config=_ConfigMock())
        _, f = tempfile.mkstemp()
        save(f, "some contents")
        with pytest.raises(InternalErrorException, match="tururu"):
            uploader.upload("fake_url", self.f, dedup=True)



import platform
import re
import textwrap

import pytest

from conan.test.utils.tools import TestClient

xcode_project = textwrap.dedent("""
    name: app
    targets:
      app:
        type: tool
        platform: macOS
        sources:
          - app
        configFiles:
          Debug: conan_config.xcconfig
          Release: conan_config.xcconfig
    """)

xcode_project_bare = textwrap.dedent("""
    name: app
    targets:
      app:
        type: tool
        platform: macOS
        sources:
          - app
    """)


main = textwrap.dedent("""
    #include <iostream>
    #include "hello.h"
    int main(int argc, char *argv[]) {
        hello();
        #ifndef DEBUG
        std::cout << "App Release!" << std::endl;
        #else
        std::cout << "App Debug!" << std::endl;
        #endif
    }
    """)

test = textwrap.dedent("""
    import os
    from conan import ConanFile, tools
    from conan.tools.build import cross_building
    class TestApp(ConanFile):
        settings = "os", "compiler", "build_type", "arch"
        def requirements(self):
            self.requires(self.tested_reference_str)
        def test(self):
            if not cross_building(self):
                self.run("app", env="conanrun")
    """)


@pytest.fixture(scope="module")
def client():
    client = TestClient()
    client.run("new cmake_lib -d name=hello -d version=0.1")
    client.run("create . -s build_type=Release")
    client.run("create . -s build_type=Debug")
    return client


@pytest.mark.skipif(platform.system() != "Darwin", reason="Only for MacOS")
@pytest.mark.tool("xcodebuild")
@pytest.mark.tool("xcodegen")
def test_project_xcodebuild(client):

    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.apple import XcodeBuild
        from conan.tools.files import copy
        class MyApplicationConan(ConanFile):
            name = "myapplication"
            version = "1.0"
            requires = "hello/0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "XcodeDeps"
            exports_sources = "app.xcodeproj/*", "app/*"
            package_type = "application"
            def build(self):
                xcode = XcodeBuild(self)
                xcode.build("app.xcodeproj")

            def package(self):
                copy(self, "build/{}/app".format(self.settings.build_type), self.source_folder,
                     os.path.join(self.package_folder, "bin"), keep_path=False)

            def package_info(self):
                self.cpp_info.bindirs = ["bin"]
        """)

    client.save({"conanfile.py": conanfile,
                 "test_package/conanfile.py": test,
                 "app/main.cpp": main,
                 "project.yml": xcode_project}, clean_first=True)
    client.run("install . --build=missing")
    client.run("install . -s build_type=Debug --build=missing")
    client.run_command("xcodegen generate")
    client.run("create . --build=missing -s os.version=15.0 -c tools.build:verbosity=verbose -c tools.compilation:verbosity=verbose")
    assert "MACOSX_DEPLOYMENT_TARGET=15.0" in client.out
    assert "xcodebuild: error: invalid option" not in client.out
    assert "hello/0.1: Hello World Release!" in client.out
    assert "App Release!" in client.out
    client.run("create . -s build_type=Debug -s os.version=15.0 --build=missing")
    assert "hello/0.1: Hello World Debug!" in client.out
    assert "App Debug!" in client.out


@pytest.mark.skipif(platform.system() != "Darwin", reason="Only for MacOS")
@pytest.mark.tool("xcodebuild")
@pytest.mark.tool("xcodegen")
@pytest.mark.skip(reason="Different sdks not installed in CI")
def test_xcodebuild_test_different_sdk(client):

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.apple import XcodeBuild
        class MyApplicationConan(ConanFile):
            name = "myapplication"
            version = "1.0"
            requires = "hello/0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "XcodeDeps"
            exports_sources = "app.xcodeproj/*", "app/*"
            def build(self):
                xcode = XcodeBuild(self)
                xcode.build("app.xcodeproj")
                self.run("otool -l build/Release/app")
        """)

    client.save({"conanfile.py": conanfile,
                 "app/main.cpp": main,
                 "project.yml": xcode_project}, clean_first=True)
    client.run("install . --build=missing")
    client.run("install . -s build_type=Debug --build=missing")
    client.run_command("xcodegen generate")
    client.run("create . --build=missing -s os.sdk=macosx -s os.sdk_version=10.15 "
               "-c tools.apple:sdk_path='/Applications/Xcode11.7.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX10.15.sdk'")
    assert "sdk 10.15.6" in client.out
    client.run("create . --build=missing -s os.sdk_version=11.3 "
               "-c tools.apple:sdk_path='/Applications/Xcode12.5.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX11.3.sdk'")
    assert "sdk 11.3" in client.out


@pytest.mark.skipif(platform.system() != "Darwin", reason="Only for MacOS")
@pytest.mark.tool("xcodebuild")
@pytest.mark.tool("xcodegen")
def test_missing_sdk(client):

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.apple import XcodeBuild
        class MyApplicationConan(ConanFile):
            name = "myapplication"
            version = "1.0"
            requires = "hello/0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "XcodeDeps"
            exports_sources = "app.xcodeproj/*", "app/*"
            def build(self):
                xcode = XcodeBuild(self)
                xcode.build("app.xcodeproj")
        """)

    client.save({"conanfile.py": conanfile,
                 "app/main.cpp": main,
                 "project.yml": xcode_project}, clean_first=True)
    client.run("install . --build=missing")
    client.run("install . -s build_type=Debug --build=missing")
    client.run_command("xcodegen generate")
    client.run("create . --build=missing -s os.sdk=macosx -s os.sdk_version=12.0 "
               "-c tools.apple:sdk_path=notexistingsdk", assert_error=True)


@pytest.mark.skipif(platform.system() != "Darwin", reason="Only for MacOS")
@pytest.mark.tool("xcodebuild")
@pytest.mark.tool("xcodegen")
@pytest.mark.parametrize("no_copy_source", [True, False])
def test_project_xcodebuild_cli_args(client, no_copy_source):

    conanfile = textwrap.dedent(f"""
        import os
        from conan import ConanFile
        from conan.tools.apple import XcodeBuild
        from conan.tools.files import copy
        class MyApplicationConan(ConanFile):
            name = "myapplication"
            version = "1.0"
            requires = "hello/0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "XcodeDeps"
            exports_sources = "app.xcodeproj/*", "app/*"
            package_type = "application"
            no_copy_source = {str(no_copy_source)}
            def build(self):
                xb = XcodeBuild(self)
                proj = os.path.join(self.source_folder, "app.xcodeproj")
                xc = os.path.join(self.build_folder, "conan_config.xcconfig")
                xb.build(proj, cli_args=["-xcconfig", xc,
                                        f"SYMROOT={{self.build_folder}}",
                                        f"OBJROOT={{self.build_folder}}"])

            def package(self):
                copy(self, "{{}}/app".format(self.settings.build_type), self.build_folder,
                     os.path.join(self.package_folder, "bin"), keep_path=False)

            def package_info(self):
                self.cpp_info.bindirs = ["bin"]
        """)

    client.save({"conanfile.py": conanfile,
                      "test_package/conanfile.py": test,
                      "app/main.cpp": main,
                      "project.yml": xcode_project_bare}, clean_first=True)

    client.run_command("xcodegen generate")

    for build_type in ["Release", "Debug"]:
        client.run(f"create . --build=missing -s build_type={build_type} -c tools.build:verbosity=verbose -c tools.compilation:verbosity=verbose")

        build_folder = re.search(r"Building your package in (/.+)", client.out).group(1)

        assert f"OBJROOT = {build_folder}"
        assert f"SYMROOT = {build_folder}"
        assert "-xcconfig" in client.out
        assert f"App {build_type}!" in client.out



import os
import textwrap

from conan.test.utils.tools import TestClient
from conan.internal.util.files import load


class TestSourceDirty:
    def test_keep_failing_source_folder(self):
        # https://github.com/conan-io/conan/issues/4025
        client = TestClient()
        conanfile = textwrap.dedent("""\
            from conan import ConanFile
            from conan.tools.files import save
            class Pkg(ConanFile):
                def source(self):
                    save(self, "somefile.txt", "hello world!!!")
                    raise Exception("boom")
            """)
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=1.0", assert_error=True)
        assert "ERROR: pkg/1.0: Error in source() method, line 6" in client.out
        # Check that we can debug and see the folder
        source_file = os.path.join(client.exported_layout().source(), "somefile.txt")
        assert load(source_file) == "hello world!!!"
        # Without any change, the export will generate same recipe revision, reuse source folder
        client.run("create . --name=pkg --version=1.0", assert_error=True)
        assert "pkg/1.0: Source folder is corrupted, forcing removal" in client.out
        assert "ERROR: pkg/1.0: Error in source() method, line 6" in client.out

        # The install also removes corrupted source folder before proceeding, then call source
        client.run("install --requires=pkg/1.0 --build=missing", assert_error=True)
        assert "pkg/1.0: WARN: Trying to remove corrupted source folder" in client.out
        assert "ERROR: pkg/1.0: Error in source() method, line 6" in client.out

        # This creates a new revision that doesn't need removal, different source folder
        client.save({"conanfile.py": conanfile.replace("source(", "source2(")})
        client.run("create . --name=pkg --version=1.0")
        assert "corrupted, forcing removal" not in client.out
        # Check that it is empty
        assert os.listdir(os.path.join(client.exported_layout().source())) == []



import os
import textwrap
from unittest.mock import patch

import pytest

from conan import conan_version
from conan.test.utils.tools import TestClient
from conan.internal.api.migrations import ClientMigrator
from conan.internal.model.version import Version
from conan.internal.util.files import save, load


@pytest.mark.parametrize(["plugin_path", "string_replace", "new_string"],
                         [("profile.py", "msvc", "EME_ESE_VC"),
                          ("compatibility/compatibility.py", "conanfile", "conian_file")])
def test_migration_profile_checker_plugin(plugin_path, string_replace, new_string):
    t = TestClient(light=True)
    # Any command that checks the package cache generates the DB
    t.run("list")
    assert os.path.exists(os.path.join(t.cache_folder, "p", "cache.sqlite3"))

    profile_plugin_path = os.path.join(t.cache_folder, "extensions", "plugins", plugin_path)
    contents = load(profile_plugin_path)

    # Let's change the version
    version_txt_file_path = os.path.join(t.cache_folder, "version.txt")
    save(version_txt_file_path, "1.0.0")
    assert os.path.exists(os.path.join(t.cache_folder, "p", "cache.sqlite3"))

    # Do a modification to the profile plugin without changing the comment
    contents = contents.replace(string_replace, new_string)
    save(profile_plugin_path, contents)

    # Trigger the migrations
    t.run("-v")
    assert f"Migration: Successfully updated {os.path.basename(plugin_path)}" in t.out
    contents = load(profile_plugin_path)
    # Our changes are removed!!!
    assert string_replace in contents
    assert new_string not in contents

    # New client, everything new
    t2 = TestClient(light=True)
    # This generates the new plugin file
    t2.run("list")

    # Do a modification to the profile plugin but changing the comment
    profile_plugin_path2 = os.path.join(t2.cache_folder, "extensions", "plugins", plugin_path)
    contents = load(profile_plugin_path2)
    contents = contents.replace(string_replace, new_string)
    contents = contents.replace("This file was generated by Conan", "This file is from ACME corp, "
                                                                    "please don't touch it.")
    save(profile_plugin_path2, contents)

    # Let's change the version
    version_txt_file_path2 = os.path.join(t2.cache_folder, "version.txt")
    save(version_txt_file_path2, "1.0.0")

    # Trigger the migrations
    t2.run("list")
    assert "Migration: Successfully updated" not in t2.out
    contents = load(profile_plugin_path2)
    # Our Changes are kept!
    assert "This file is from ACME corp, " in contents
    assert string_replace not in contents
    assert new_string in contents


def test_back_migrations():
    t = TestClient(light=True)

    # add 3 migrations
    for number in (1, 2, 3):
        migration_file = os.path.join(t.cache_folder, "migrations", f"2.100.0_{number}-migrate.py")
        migrate = textwrap.dedent(f"""
            import os
            def migrate(cache_folder):
                os.remove(os.path.join(cache_folder, "file{number}.txt"))
            """)
        save(migration_file, migrate)
        save(os.path.join(t.cache_folder, f"file{number}.txt"), "some content")
        # Some older versions migrations that shouldn't be applied if we downgrade to current
        wrong_migration_file = os.path.join(t.cache_folder, "migrations", f"2.0_{number}-migrate.py")
        save(wrong_migration_file, "this is not python, it would crash")

    # Let's change the old version
    version_txt_file_path = os.path.join(t.cache_folder, "version.txt")
    save(version_txt_file_path, "200.0")
    t.run("-v")  # Fire the backward migration
    assert f"WARN: Downgrading cache from Conan 200.0 to {conan_version}" in t.out
    for number in (1, 2, 3):
        assert f"WARN: Applying downgrade migration 2.100.0_{number}-migrate.py" in t.out
        assert not os.path.exists(os.path.join(t.cache_folder, f"file{number}.txt"))
        migration_file = os.path.join(t.cache_folder, "migrations", f"2.100.0_{number}-migrate.py")
        assert not os.path.exists(migration_file)


def test_back_default_compatibility_migration():
    t = TestClient(light=True)
    t.run("-v")  # Fire the backward migration
    migration_file = os.path.join(t.cache_folder, "migrations", "2.4_1-migrate.py")
    assert os.path.exists(migration_file)

    # downgrade from a clean latest conan_version to 2.3.2
    # simulate that we are in 2.3.2 and the old one is latest conan_version
    migrator = ClientMigrator(t.cache_folder, Version("2.3.2"))

    with patch('conan.api.conan_api.ClientMigrator', new=lambda *args, **kwargs: migrator):
        t.run("-v")  # Fire the backward migration
        assert f"WARN: Downgrading cache from Conan {conan_version} to 2.3.2" in t.out


class TestMigrationCppstdCompat:
    def test_migration(self):
        t = TestClient(light=True)
        t.run("-v")
        cppstd_compat_path = "extensions/plugins/compatibility/cppstd_compat.py"
        compatibility_path = "extensions/plugins/compatibility/compatibility.py"

        # both files exist and not modified
        t.save_home({"version.txt": "2.11"})
        t.run("-v")
        assert "def cppstd_compat(conanfile)" in t.load_home(compatibility_path)
        assert not os.path.exists(os.path.join(t.cache_folder, cppstd_compat_path))

    def test_cppstd_modified(self):
        t = TestClient(light=True)
        t.run("-v")
        cppstd_compat_path = "extensions/plugins/compatibility/cppstd_compat.py"
        compatibility_path = "extensions/plugins/compatibility/compatibility.py"
        # cppstd_compat modified
        t.save_home({"version.txt": "2.11",
                     compatibility_path: "# This file was generated by Conan",
                     cppstd_compat_path: "custom file content"})
        t.run("-v")
        assert t.load_home(cppstd_compat_path) == "custom file content"
        # compatibility not migrated, keeps the old content
        assert "def cppstd_compat(conanfile)" not in t.load_home(compatibility_path)

    def test_compatibility_modified(self):
        t = TestClient(light=True)
        t.run("-v")
        cppstd_compat_path = "extensions/plugins/compatibility/cppstd_compat.py"
        compatibility_path = "extensions/plugins/compatibility/compatibility.py"
        t.save_home({"version.txt": "2.11",
                     cppstd_compat_path: "# This file was generated by Conan POTATO",
                     compatibility_path: "Modified file"})
        t.run("-v")
        assert t.load_home(compatibility_path) == "Modified file"
        # not Removed because compatibility was modified
        assert "POTATO" in t.load_home(cppstd_compat_path)

    def test_compatibility_modified_by_conan(self):
        t = TestClient(light=True)
        t.run("-v")
        compatibility_path = "extensions/plugins/compatibility/compatibility.py"
        old = t.load_home(compatibility_path)
        # In old versions of Conan, assume POTATO was present in the file
        new = old + "\n# POTATO"
        t.save_home({"version.txt": "2.11",
                     compatibility_path: new})
        # But now with the newer version, Conan removed it
        # so the triggered migration should remove it
        t.run("-v")
        assert "Successfully updated compatibility.py" in t.out
        assert "# POTATO" not in t.load_home(compatibility_path)

    def test_compatibility_no_rewrite(self):
        t = TestClient(light=True)
        t.run("-v")
        # compatibility.py exists and is not modified
        t.save_home({"version.txt": "2.11"})
        t.run("-v")
        assert "Successfully updated compatibility.py" not in t.out



import os
import platform
import textwrap

import pytest

from conan.api.model import PkgReference
from conan.api.model import RecipeReference
from conan.internal.paths import PACKAGE_TGZ_NAME
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient, TestServer
from conan.internal.util.files import load, rmdir, chdir, save

links_conanfile = textwrap.dedent("""
    from conan import ConanFile
    from conan.tools.files import copy

    class HelloConan(ConanFile):
        name = "hello"
        version = "0.1"
        exports_sources = "*"

        def package(self):
            copy(self, "*", self.build_folder, self.package_folder)
    """)


@pytest.mark.skipif(platform.system() == "Windows", reason="Requires Symlinks")
def test_complete_round_trip():
    """ a full round trip with a in package symlink, that is maintained all the way to
    the server and re-deployed in the client
    the symlink is relative just "link.txt" -> "midlink.txt" -> "target.txt" (local)
    """
    c = TestClient(default_server_user=True)
    c.save({"conanfile.py": links_conanfile,
            "target.txt": "hello world!"})
    os.symlink("target.txt", os.path.join(c.current_folder, "midlink.txt"))
    os.symlink("midlink.txt", os.path.join(c.current_folder, "link.txt"))
    assert c.load("link.txt") == "hello world!"

    c.run("create .")

    def checker(folder):
        with chdir(folder):
            assert os.path.exists("target.txt")
            assert os.readlink("link.txt") == "midlink.txt"
            assert os.readlink("midlink.txt") == "target.txt"
            assert load("link.txt") == "hello world!"

    ref = RecipeReference.loads("hello/0.1")
    checker(c.get_latest_ref_layout(ref).source())
    pref = c.get_latest_package_reference(ref)
    pkg_layout = c.get_latest_pkg_layout(pref)
    checker(pkg_layout.build())
    checker(pkg_layout.package())

    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.save({}, clean_first=True)
    c.run("install --requires=hello/0.1 --deployer=full_deploy")
    checker(os.path.join(c.current_folder, "full_deploy", "host", "hello", "0.1"))


@pytest.mark.skipif(platform.system() == "Windows", reason="Requires Symlinks")
def test_complete_round_trip_broken_link():
    """ same as above but with a broken one
    link.txt->midlink.txt->(broken)
    """
    c = TestClient(default_server_user=True)
    c.save({"conanfile.py": links_conanfile})
    os.symlink("target.txt", os.path.join(c.current_folder, "midlink.txt"))
    os.symlink("midlink.txt", os.path.join(c.current_folder, "link.txt"))

    c.run("create .")

    def checker(folder):
        with chdir(folder):
            assert not os.path.exists("target.txt")
            assert os.readlink("link.txt") == "midlink.txt"
            assert os.readlink("midlink.txt") == "target.txt"

    ref = RecipeReference.loads("hello/0.1")
    checker(c.get_latest_ref_layout(ref).source())
    pref = c.get_latest_package_reference(ref)
    pkg_layout = c.get_latest_pkg_layout(pref)
    checker(pkg_layout.build())
    checker(pkg_layout.package())

    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.save({}, clean_first=True)
    c.run("install --requires=hello/0.1 --deployer=full_deploy")
    checker(os.path.join(c.current_folder, "full_deploy", "host", "hello", "0.1"))


@pytest.mark.skipif(platform.system() == "Windows", reason="Requires Symlinks")
def test_complete_round_trip_external_link():
    """ same as above but with a broken one
    link.txt->midlink.txt->/abs/path/to/target.txt
    """
    c = TestClient(default_server_user=True)
    target = os.path.join(temp_folder(), "target.txt")
    save(target, "foo")

    c.save({"conanfile.py": links_conanfile})
    os.symlink(target, os.path.join(c.current_folder, "midlink.txt"))
    os.symlink("midlink.txt", os.path.join(c.current_folder, "link.txt"))

    c.run("create .")

    def checker(folder):
        with chdir(folder):
            assert "target.txt" not in os.listdir(".")
            assert not os.path.exists("target.txt")
            assert os.readlink("link.txt") == "midlink.txt"
            assert os.readlink("midlink.txt") == target
            assert load("link.txt") == "foo"

    ref = RecipeReference.loads("hello/0.1")
    checker(c.get_latest_ref_layout(ref).source())
    pref = c.get_latest_package_reference(ref)
    pkg_layout = c.get_latest_pkg_layout(pref)
    checker(pkg_layout.build())
    checker(pkg_layout.package())

    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.save({}, clean_first=True)
    c.run("install --requires=hello/0.1 --deployer=full_deploy")
    checker(os.path.join(c.current_folder, "full_deploy", "host", "hello", "0.1"))


@pytest.mark.skipif(platform.system() == "Windows", reason="Requires Symlinks")
def test_complete_round_trip_folders():
    """ similar to above, but with 2 folder symlinks and one file symlink
    # Reproduces issue: https://github.com/conan-io/conan/issues/5329
    """
    c = TestClient(default_server_user=True)

    c.save({"conanfile.py": links_conanfile,
            "src/framework/Versions/v1/headers/content": "myheader!",
            "src/framework/Versions/v1/file": "myfile!"})

    # Add two levels of symlinks
    os.symlink('v1', os.path.join(c.current_folder, 'src', 'framework', 'Versions', 'Current'))
    os.symlink('Versions/Current/headers',
               os.path.join(c.current_folder, 'src', 'framework', 'headers'))
    os.symlink('Versions/Current/file',
               os.path.join(c.current_folder, 'src', 'framework', 'file'))

    c.run("create .")

    def checker(folder):
        with chdir(folder):
            assert os.readlink("src/framework/Versions/Current") == "v1"
            assert os.readlink("src/framework/headers") == "Versions/Current/headers"
            assert os.readlink("src/framework/file") == "Versions/Current/file"
            assert os.path.exists("src/framework/Versions/v1/headers/content")
            assert os.path.exists("src/framework/Versions/v1/file")

            assert load("src/framework/file") == "myfile!"
            assert load("src/framework/headers/content") == "myheader!"

    ref = RecipeReference.loads("hello/0.1")
    checker(c.get_latest_ref_layout(ref).source())
    pref = c.get_latest_package_reference(ref)
    pkg_layout = c.get_latest_pkg_layout(pref)
    checker(pkg_layout.build())
    checker(pkg_layout.package())

    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.save({}, clean_first=True)
    c.run("install --requires=hello/0.1 --deployer=full_deploy")
    checker(os.path.join(c.current_folder, "full_deploy", "host", "hello", "0.1"))


@pytest.mark.skipif(platform.system() != "Linux", reason="Only linux")
@pytest.mark.parametrize("package_files",
                         [{"files": ["foo/bar/folder/file.txt",
                                     "foo/bar/folder/other/other_file.txt"],
                           "symlinks": [("../file.txt", "foo/bar/folder/other/file2.txt")]},
                          # relative ../ symlink
                          {"files": ["foo/bar/file/file.txt"],
                           "symlinks": [(temp_folder(), "foo/symlink_folder")]},  # absolute symlink
                          {"files": ["folder/file.txt"],
                           "symlinks": [("folder", "folder2"),
                                        ("file.txt", "folder/file2.txt")]},  # single level symlink
                          {"files": ["foo/bar/file/file.txt"],
                           "symlinks": [("bar/file", "foo/symlink_folder"),
                                        ("foo/symlink_folder/file.txt", "file2.txt")]},
                          # double level symlink
                          ])
def test_package_with_symlinks(package_files):

    client = TestClient(default_server_user=True)
    client2 = TestClient(servers=client.servers)
    client.save({"conanfile.py": links_conanfile})

    for path in package_files["files"]:
        client.save({path: "foo contents"})

    for link_dest, link_file in package_files["symlinks"]:
        os.symlink(link_dest, os.path.join(client.current_folder, link_file))

    client.run("create .")

    def assert_folder_symlinks(base_folder):
        with chdir(base_folder):
            for f in package_files["files"]:
                assert os.path.exists(f)
            for link_dst, link in package_files["symlinks"]:
                assert os.readlink(link) == link_dst
                if os.path.isfile(link):
                    assert load(link) == "foo contents"

    # Check exported sources are there
    ref_layout = client.exported_layout()
    assert_folder_symlinks(ref_layout.export_sources())
    assert_folder_symlinks(ref_layout.source())

    # Check files have been copied to the build
    pkg_layout = client.created_layout()
    assert_folder_symlinks(pkg_layout.build())
    assert_folder_symlinks(pkg_layout.package())

    client.run("upload '*' -c -r default")
    client.run("remove * -c")

    # Client 2 install
    client2.run("install --requires=hello/0.1 --deployer=full_deploy")
    ref = client2.get_latest_ref_layout(RecipeReference.loads("hello/0.1"))
    pref = PkgReference(ref.reference, "da39a3ee5e6b4b0d3255bfef95601890afd80709")

    # Check package files are there
    package_folder = client2.get_latest_pkg_layout(pref).package()
    assert_folder_symlinks(package_folder)
    assert_folder_symlinks(os.path.join(client2.current_folder, "full_deploy", "host", "hello", "0.1"))


@pytest.mark.skipif(platform.system() == "Windows", reason="Symlinks not in Windows")
def test_exports_does_not_follow_symlink():
    tmp = temp_folder()
    linked_abs_folder = tmp
    save(os.path.join(tmp, "source.cpp"), "foo")
    client = TestClient(default_server_user=True)
    conanfile = GenConanfile("lib", "1.0")\
        .with_package('copy(self, "*", self.source_folder, self.package_folder)')\
        .with_exports_sources("*")\
        .with_import("from conan.tools.files import copy")
    client.save({"conanfile.py": conanfile, "foo.txt": "bar"})
    os.symlink(linked_abs_folder, os.path.join(client.current_folder, "linked_folder"))
    client.run("create . ")
    exports_sources_folder = client.exported_layout().export_sources()
    assert os.path.islink(os.path.join(exports_sources_folder, "linked_folder"))
    assert os.path.exists(os.path.join(exports_sources_folder, "linked_folder", "source.cpp"))

    # Check files have been copied to the build
    build_folder = client.created_layout().build()
    assert os.path.islink(os.path.join(build_folder, "linked_folder"))
    assert os.path.exists(os.path.join(build_folder, "linked_folder", "source.cpp"))

    # Check package files are there
    package_folder = client.created_layout().package()
    assert os.path.islink(os.path.join(package_folder, "linked_folder"))
    assert os.path.exists(os.path.join(package_folder, "linked_folder", "source.cpp"))

    # Check that the manifest doesn't contain the symlink nor the source.cpp
    contents = load(os.path.join(package_folder, "conanmanifest.txt"))
    assert "foo.txt" in contents
    assert "linked_folder" not in contents
    assert "source.cpp" not in contents

    # Now is a broken link, but the files are not in the cache, just a broken link
    rmdir(linked_abs_folder)
    assert not os.path.exists(os.path.join(exports_sources_folder, "linked_folder", "source.cpp"))
    assert not os.path.exists(os.path.join(build_folder, "linked_folder", "source.cpp"))
    assert not os.path.exists(os.path.join(package_folder, "linked_folder", "source.cpp"))


@pytest.mark.skipif(platform.system() != "Linux", reason="Only linux")
def test_package_symlinks_zero_size():
    server = TestServer()
    client = TestClient(servers={"default": server}, inputs=["admin", "password"])

    conanfile = """
import os
from conan import ConanFile
from conan.tools.files import save

class HelloConan(ConanFile):
    name = "hello"
    version = "0.1"

    def package(self):
        # Link to file.txt and then remove it
        save(self, os.path.join(self.package_folder, "file.txt"), "contents")
        os.symlink("file.txt", os.path.join(self.package_folder, "link.txt"))
"""
    client.save({"conanfile.py": conanfile})
    # By default it is not allowed
    client.run("create .")
    p_folder = client.created_layout().download_package()
    # Upload, it will create the tgz
    client.run("upload * -r=default -c")

    # We can uncompress it without warns
    tgz = os.path.join(p_folder, PACKAGE_TGZ_NAME)
    client.run_command('gzip -d "{}"'.format(tgz))
    client.run_command('tar tvf "{}"'.format(os.path.join(p_folder, "conan_package.tar")))
    lines = str(client.out).splitlines()
    """
-rw-r--r-- 0/0               8 1970-01-01 01:00 file.txt
lrw-r--r-- 0/0               0 1970-01-01 01:00 link.txt -> file.txt
    """

    assert "link.txt" in " ".join(lines)
    for line in lines:
        assert ".txt" in line
        size = int([i for i in line.split(" ") if i][2])
        if "link.txt" in line:
            assert int(size) == 0
        elif "file.txt":
            assert int(size) > 0
